import os
import math
import csv
import torch 
import numpy as np
import pandas as pd
from skimage.feature import peak_local_max, blob_log
from skimage.measure import label, regionprops_table
from skimage.morphology import disk, dilation        
from tqdm import tqdm

from biapy.data.data_2D_manipulation import load_and_prepare_2D_train_data
from biapy.data.data_3D_manipulation import load_and_prepare_3D_data
from biapy.data.post_processing.post_processing import (remove_close_points, detection_watershed, 
    measure_morphological_props_and_filter)
from biapy.data.pre_processing import create_detection_masks, norm_range01
from biapy.utils.util import save_tif, read_chunked_data, write_chunked_data, order_dimensions
from biapy.engine.metrics import detection_metrics, jaccard_index, weighted_bce_dice_loss, CrossEntropyLoss_wrapper
from biapy.engine.base_workflow import Base_Workflow

class Detection_Workflow(Base_Workflow):
    """
    Detection workflow where the goal is to localize objects in the input image, not requiring a pixel-level class.
    More details in `our documentation <https://biapy.readthedocs.io/en/latest/workflows/detection.html>`_.  

    Parameters
    ----------
    cfg : YACS configuration
        Running configuration.
    
    Job_identifier : str
        Complete name of the running job.

    device : Torch device
        Device used. 

    args : argpase class
        Arguments used in BiaPy's call. 
    """
    def __init__(self, cfg, job_identifier, device, args, **kwargs):
        super(Detection_Workflow, self).__init__(cfg, job_identifier, device, args, **kwargs)

        # Detection stats
        self.stats['d_precision'] = 0
        self.stats['d_recall'] = 0
        self.stats['d_f1'] = 0

        self.stats['d_precision_merge_patches'] = 0
        self.stats['d_recall_merge_patches'] = 0
        self.stats['d_f1_merge_patches'] = 0
        
        self.original_test_mask_path = self.prepare_detection_data()

        self.use_gt = False 
        if self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            self.use_gt = True 
        if self.cfg.TEST.BY_CHUNKS.ENABLE and self.cfg.TEST.BY_CHUNKS.WORKFLOW_PROCESS.ENABLE:
            self.use_gt = False 

        if self.use_gt:
            self.csv_files = sorted(next(os.walk(self.original_test_mask_path))[2])
        self.cell_count_file = os.path.join(self.cfg.PATHS.RESULT_DIR.PATH, 'cell_counter.csv')
        self.cell_count_lines = []

        # From now on, no modification of the cfg will be allowed
        self.cfg.freeze()

        # Activations for each output channel:
        # channel number : 'activation'
        self.activations = [{'0': 'CE_Sigmoid'}]

        # Workflow specific training variables
        self.mask_path = cfg.DATA.TRAIN.GT_PATH
        self.load_Y_val = True

        # Workflow specific test variables
        self.postpone_postproc = False
        if cfg.TEST.BY_CHUNKS.ENABLE and cfg.TEST.BY_CHUNKS.WORKFLOW_PROCESS.ENABLE and \
            cfg.TEST.BY_CHUNKS.WORKFLOW_PROCESS.TYPE == "chunk_by_chunk":
            self.postpone_postproc = True

        if self.cfg.TEST.POST_PROCESSING.DET_WATERSHED or self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS:
            self.post_processing['detection_post'] = True
        else:
            self.post_processing['detection_post'] = False    

    def define_metrics(self):
        """
        Definition of self.metrics, self.metric_names and self.loss variables.
        """
        self.metrics = [
            jaccard_index(num_classes=self.cfg.MODEL.N_CLASSES, device=self.device, 
                torchvision_models=True if self.cfg.MODEL.SOURCE == "torchvision" else False)
        ]
        self.metric_names = ["jaccard_index"]
        if self.cfg.LOSS.TYPE == "CE": 
            self.loss = CrossEntropyLoss_wrapper(num_classes=self.cfg.MODEL.N_CLASSES,
                torchvision_models=True if self.cfg.MODEL.SOURCE == "torchvision" else False)
        elif self.cfg.LOSS.TYPE == "W_CE_DICE":
            self.loss = weighted_bce_dice_loss(w_dice=0.66, w_bce=0.33)

    def metric_calculation(self, output, targets, metric_logger=None):
        """
        Execution of the metrics defined in :func:`~define_metrics` function. 

        Parameters
        ----------
        output : Torch Tensor
            Prediction of the model. 

        targets : Torch Tensor
            Ground truth to compare the prediction with. 

        metric_logger : MetricLogger, optional
            Class to be updated with the new metric(s) value(s) calculated. 
        
        Returns
        -------
        value : float
            Value of the metric for the given prediction. 
        """
        with torch.no_grad():
            train_iou = self.metrics[0](output, targets)
            train_iou = train_iou.item() if not torch.isnan(train_iou) else 0
            if metric_logger is not None:
                metric_logger.meters[self.metric_names[0]].update(train_iou)
            else:
                return train_iou

    def detection_process(self, pred, filenames, metric_names=[]):
        """
        Detection workflow engine for test/inference. Process model's prediction to prepare detection output and 
        calculate metrics. 

        Parameters
        ----------
        pred : Torch Tensor
            Model predictions.
        
        filenames : List of str
            Predicted image's filenames.

        metric_names : List of str
            Metrics names.
        """
        file_ext = os.path.splitext(filenames[0])[1]
        ndim = 2 if self.cfg.PROBLEM.NDIM == "2D" else 3
        pred_shape = pred.shape
        print("Capturing the local maxima ")
        all_points = []
        all_classes = []
        for channel in range(pred.shape[-1]):
            print("Class {}".format(channel+1))
            if len(self.cfg.TEST.DET_MIN_TH_TO_BE_PEAK) == 1:
                min_th_peak = self.cfg.TEST.DET_MIN_TH_TO_BE_PEAK[0]
            else:
                min_th_peak = self.cfg.TEST.DET_MIN_TH_TO_BE_PEAK[channel]
            
            # Find points
            if self.cfg.TEST.DET_POINT_CREATION_FUNCTION == "peak_local_max":
                pred_coordinates = peak_local_max(pred[...,channel].astype(np.float32),
                    min_distance=self.cfg.TEST.DET_PEAK_LOCAL_MAX_MIN_DISTANCE, threshold_abs=min_th_peak,
                    exclude_border=self.cfg.TEST.DET_EXCLUDE_BORDER)
            else:
                pred_coordinates = blob_log(pred[...,channel]*255, min_sigma=self.cfg.TEST.DET_BLOB_LOG_MIN_SIGMA, 
                    max_sigma=self.cfg.TEST.DET_BLOB_LOG_MAX_SIGMA, num_sigma=self.cfg.TEST.DET_BLOB_LOG_NUM_SIGMA, 
                    threshold=min_th_peak, exclude_border=self.cfg.TEST.DET_EXCLUDE_BORDER)
                pred_coordinates = pred_coordinates[:,:3].astype(int) # Remove sigma

            # Remove close points per class as post-processing method
            if self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS and not self.postpone_postproc:
                if len(self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS) == 1:
                    radius = self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS[0]
                else:
                    radius = self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS[channel]

                pred_coordinates = remove_close_points(pred_coordinates, radius, self.cfg.DATA.TEST.RESOLUTION,
                    ndim=ndim)
                    
            all_points.append(pred_coordinates)   
            c_size = 1 if len(pred_coordinates) == 0 else len(pred_coordinates)
            all_classes.append(np.full(c_size, channel))

        # Remove close points again seeing all classes together, as it can be that a point is detected in both classes
        # if there is not clear distinction between them
        classes = 1 if self.cfg.MODEL.N_CLASSES <= 2 else self.cfg.MODEL.N_CLASSES
        if self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS and classes > 1 and not self.postpone_postproc:
            print("All classes together")
            radius = np.min(self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS)

            all_points = np.concatenate(all_points, axis=0)
            all_classes = np.concatenate(all_classes, axis=0)

            new_points, all_classes = remove_close_points(all_points, radius, self.cfg.DATA.TEST.RESOLUTION,
                classes=all_classes, ndim=ndim)
            
            # Create again list of arrays of all points
            all_points = []
            for i in range(classes):
                all_points.append([])
            for i, c in enumerate(all_classes):
                all_points[c].append(new_points[i])
            del new_points
        # Create a file with detected point and other image with predictions ids (if GT given)
        print("Creating the images with detected points . . .")   
        points_pred = np.zeros(pred.shape[:-1], dtype=np.uint8)
        for n, pred_coordinates in enumerate(all_points):
            if self.use_gt:
                pred_id_img = np.zeros(pred_shape[:-1], dtype=np.uint32)
            for j, coord in enumerate(pred_coordinates):
                z,y,x = coord
                points_pred[z,y,x] = n+1
                if self.use_gt:
                    pred_id_img[z,y,x] = j+1
            
            # Dilate and save the prediction ids for the current class 
            if self.use_gt:
                for i in range(pred_id_img.shape[0]):                                                                                  
                    pred_id_img[i] = dilation(pred_id_img[i], disk(3))
                if file_ext in ['.hdf5', '.h5', ".zarr"]:
                    write_chunked_data(np.expand_dims(pred_id_img,-1), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, 
                        os.path.splitext(filenames[0])[0]+'_class'+str(n+1)+'_pred_ids'+file_ext, dtype_str="uint32", 
                        verbose=self.cfg.TEST.VERBOSE)
                else:
                    save_tif(np.expand_dims(np.expand_dims(pred_id_img,0),-1), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS,
                        [os.path.splitext(filenames[0])[0]+'_class'+str(n+1)+'_pred_ids.tif'], verbose=self.cfg.TEST.VERBOSE)

            self.cell_count_lines.append([filenames, len(pred_coordinates)])

        if self.use_gt: del pred_id_img

        # Dilate and save the detected point image
        if len(pred_coordinates) > 0:
            for i in range(points_pred.shape[0]):                                                                                  
                points_pred[i] = dilation(points_pred[i], disk(3)) 
        if file_ext in ['.hdf5', '.h5', ".zarr"]:
            write_chunked_data(np.expand_dims(points_pred,-1), self.cfg.PATHS.RESULT_DIR.DET_LOCAL_MAX_COORDS_CHECK, filenames[0], 
                dtype_str="uint8", verbose=self.cfg.TEST.VERBOSE)
        else:
            save_tif(np.expand_dims(np.expand_dims(points_pred,0),-1), self.cfg.PATHS.RESULT_DIR.DET_LOCAL_MAX_COORDS_CHECK,
                filenames, verbose=self.cfg.TEST.VERBOSE)

        # Detection watershed
        if self.cfg.TEST.POST_PROCESSING.DET_WATERSHED:
            data_filename = os.path.join(self.cfg.DATA.TEST.PATH, filenames[0])
            w_dir = os.path.join(self.cfg.PATHS.WATERSHED_DIR, filenames[0])
            check_wa = w_dir if self.cfg.PROBLEM.DETECTION.DATA_CHECK_MW else None
            points_pred = detection_watershed(points_pred, all_points, data_filename, self.cfg.TEST.POST_PROCESSING.DET_WATERSHED_FIRST_DILATION,
                clases, ndim=ndim, donuts_classes=self.cfg.TEST.POST_PROCESSING.DET_WATERSHED_DONUTS_CLASSES,
                donuts_patch=self.cfg.TEST.POST_PROCESSING.DET_WATERSHED_DONUTS_PATCH, 
                donuts_nucleus_diameter=self.cfg.TEST.POST_PROCESSING.DET_WATERSHED_DONUTS_NUCLEUS_DIAMETER, save_dir=check_wa)
            
            # Instance filtering by properties     
            points_pred, d_result = measure_morphological_props_and_filter(points_pred, self.cfg.DATA.TEST.RESOLUTION, 
                properties=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.PROPS, 
                prop_values=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.VALUES, 
                comp_signs=self.cfg.TEST.POST_PROCESSING.MEASURE_PROPERTIES.REMOVE_BY_PROPERTIES.SIGN, 
                coords_list=np.concatenate(all_points, axis=0))

            if file_ext in ['.hdf5', '.h5', ".zarr"]:
                write_chunked_data(np.expand_dims(points_pred,-1), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, filenames[0], dtype_str="uint8", 
                    verbose=self.cfg.TEST.VERBOSE)
            else:
                save_tif(np.expand_dims(np.expand_dims(points_pred,0),-1), self.cfg.PATHS.RESULT_DIR.PER_IMAGE_POST_PROCESSING,
                    filenames, verbose=self.cfg.TEST.VERBOSE)
        del points_pred

        # Save coords in a couple of csv files            
        aux = np.concatenate(all_points, axis=0)
        df = None
        if len(aux) != 0:
            if self.cfg.PROBLEM.NDIM == "3D":
                prob = pred[aux[:,0], aux[:,1], aux[:,2], all_classes]
                prob = np.concatenate(prob, axis=0)
                all_classes = np.concatenate(all_classes, axis=0)
                if self.cfg.TEST.POST_PROCESSING.DET_WATERSHED:
                    df = pd.DataFrame(zip(d_result['labels'], list(aux[:,0]), list(aux[:,1]), list(aux[:,2]), list(prob), list(all_classes),
                        d_result['npixels'], d_result['areas'], d_result['circularities'], d_result['diameters'], d_result['perimeters'], 
                        d_result['comment'], d_result['conditions']), columns =['pred_id', 'axis-0', 'axis-1', 'axis-2', 'probability', 
                        'class', 'npixels', 'volume', 'sphericity', 'diameter', 'perimeter (surface area)', 'comment', 'conditions'])
                    df = df.sort_values(by=['pred_id'])   
                else:
                    labels = []
                    for i, pred_coordinates in enumerate(all_points):
                        for j in range(len(pred_coordinates)):
                            labels.append(j+1)

                    df = pd.DataFrame(zip(labels, list(aux[:,0]), list(aux[:,1]), list(aux[:,2]), list(prob), list(all_classes)), 
                        columns =['pred_id', 'axis-0', 'axis-1', 'axis-2', 'probability', 'class'])
                    df = df.sort_values(by=['pred_id'])
            else:
                aux = aux[:,1:]
                prob = pred[0,aux[:,0], aux[:,1], all_classes]
                prob = np.concatenate(prob, axis=0)
                all_classes = np.concatenate(all_classes, axis=0)
                if self.cfg.TEST.POST_PROCESSING.DET_WATERSHED:
                    df = pd.DataFrame(zip(d_result['labels'], list(aux[:,0]), list(aux[:,1]), list(prob), list(all_classes),
                        d_result['npixels'], d_result['areas'], d_result['circularities'], d_result['diameters'], d_result['perimeters'], 
                        d_result['elongations'], d_result['comment'], d_result['conditions']), columns =['pred_id', 'axis-0', 'axis-1', 
                        'probability', 'class', 'npixels', 'area', 'circularity', 'diameter', 'perimeter', 'elongation', 'comment', 
                        'conditions'])
                    df = df.sort_values(by=['pred_id'])   
                else:
                    df = pd.DataFrame(zip(list(aux[:,0]), list(aux[:,1]), list(prob), list(all_classes)), 
                        columns =['axis-0', 'axis-1', 'probability', 'class'])
                    df = df.sort_values(by=['axis-0'])
            del aux 

            # Save jus the points and their probabilities 
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.DET_LOCAL_MAX_COORDS_CHECK, os.path.splitext(filenames[0])[0]+'_full_info.csv'))
            if self.cfg.TEST.POST_PROCESSING.DET_WATERSHED:
                if ndim == 2:
                    cols = ['class', 'pred_id', 'npixels', 'area', 'circularity', 'perimeter', 'elongation', 'comment', 'conditions']
                else:
                    cols = ['class', 'pred_id', 'npixels', 'volume', 'sphericity', 'perimeter', 'surface area', 'comment', 'conditions']
                df = df.drop(columns=cols)
            else:
                df = df.drop(columns=['class'])
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.DET_LOCAL_MAX_COORDS_CHECK, os.path.splitext(filenames[0])[0]+'_prob.csv'))

        # Calculate detection metrics
        if self.use_gt:
            all_channel_d_metrics = [0,0,0]
            dfs = []
            gt_all_coords = []
            for ch, pred_coordinates in enumerate(all_points):

                # Read the GT coordinates from the CSV file
                csv_filename = os.path.join(self.original_test_mask_path, os.path.splitext(filenames[0])[0]+'.csv')
                if not os.path.exists(csv_filename):
                    print("WARNING: The CSV file seems to have different name than image. Using the CSV file "
                            "with the same position as the CSV in the directory. Check if it is correct!")
                    csv_filename = os.path.join(self.original_test_mask_path, self.csv_files[self.f_numbers[0]])
                    print("Its respective CSV file seems to be: {}".format(csv_filename))
                print("Reading GT data from: {}".format(csv_filename))
                df_gt = pd.read_csv(csv_filename, index_col=0)     
                zcoords = df_gt['axis-0'].tolist()
                ycoords = df_gt['axis-1'].tolist()
                if self.cfg.PROBLEM.NDIM == '3D': 
                    xcoords = df_gt['axis-2'].tolist()
                    gt_coordinates = [[z,y,x] for z,y,x in zip(zcoords,ycoords,xcoords)]
                else:
                    gt_coordinates = [[0,y,x] for y,x in zip(zcoords,ycoords)]
                gt_all_coords.append(gt_coordinates)

                if self.cfg.PROBLEM.NDIM == '3D':
                    v_size = (self.cfg.DATA.TEST.RESOLUTION[0], self.cfg.DATA.TEST.RESOLUTION[1], self.cfg.DATA.TEST.RESOLUTION[2])
                else:
                    v_size = (1,self.cfg.DATA.TEST.RESOLUTION[0], self.cfg.DATA.TEST.RESOLUTION[1])

                # Calculate detection metrics 
                if len(pred_coordinates) > 0:
                    print("Detection (class "+str(ch+1)+")")
                    d_metrics, gt_assoc, fp = detection_metrics(gt_coordinates, pred_coordinates, tolerance=self.cfg.TEST.DET_TOLERANCE[ch],
                        voxel_size=v_size, return_assoc=True, verbose=self.cfg.TEST.VERBOSE)
                    print("Detection metrics: {}".format(d_metrics))
                    all_channel_d_metrics[0] += d_metrics[1]
                    all_channel_d_metrics[1] += d_metrics[3]
                    all_channel_d_metrics[2] += d_metrics[5]

                    # Save csv files with the associations between GT points and predicted ones 
                    dfs.append([gt_assoc.copy(),fp.copy()])
                    if self.cfg.PROBLEM.NDIM == "2D":
                        gt_assoc = gt_assoc.drop(columns=['axis-0'])
                        fp = fp.drop(columns=['axis-0'])
                        gt_assoc = gt_assoc.rename(columns={'axis-1': 'axis-0', 'axis-2': 'axis-1'})
                        fp = fp.rename(columns={'axis-1': 'axis-0', 'axis-2': 'axis-1'})
                    gt_assoc.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_class'+str(ch+1)+'_gt_assoc.csv'))
                    fp.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, os.path.splitext(filenames[0])[0]+'_class'+str(ch+1)+'_fp.csv'))             
                else:
                    print("No point found to calculate the metrics!")

            print("All classes "+str(ch+1))
            all_channel_d_metrics[0] = all_channel_d_metrics[0]/len(all_points)
            all_channel_d_metrics[1] = all_channel_d_metrics[1]/len(all_points)
            all_channel_d_metrics[2] = all_channel_d_metrics[2]/len(all_points)
            print("Detection metrics: {}".format(["Precision", all_channel_d_metrics[0],
                "Recall", all_channel_d_metrics[1], "F1", all_channel_d_metrics[2]]))

            self.stats[metric_names[0]] += all_channel_d_metrics[0]
            self.stats[metric_names[1]] += all_channel_d_metrics[1]
            self.stats[metric_names[2]] += all_channel_d_metrics[2]
            
            print("Creating the image with a summary of detected points and false positives with colors . . .")
            points_pred = np.zeros(pred_shape[:-1]+(3,), dtype=np.uint8)
            for ch, gt_coords in enumerate(gt_all_coords):
                if len(dfs) > 0:
                    gt_assoc, fp = dfs[ch]

                # TP and FN
                gt_id_img = np.zeros(pred_shape[:-1], dtype=np.uint32)
                for j, cor in enumerate(gt_coords):
                    z,y,x = cor
                    z,y,x = int(z),int(y),int(x)
                    if len(dfs) > 0:
                        if gt_assoc[gt_assoc['gt_id'] == j+1]["tag"].iloc[0] == "TP":
                            points_pred[z,y,x] = (0,255,0)# Green
                        else:   
                            points_pred[z,y,x] = (255,0,0)# Red
                    else:                           
                        points_pred[z,y,x] = (255,0,0)# Red

                    gt_id_img[z,y,x] = j+1

                # Dilate and save the GT ids for the current class 
                for i in range(gt_id_img.shape[0]):      
                    gt_id_img[i] = dilation(gt_id_img[i], disk(3))
                if file_ext in ['.hdf5', '.h5', ".zarr"]:
                    write_chunked_data(np.expand_dims(gt_id_img,-1), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, 
                        os.path.splitext(filenames[0])[0]+'_class'+str(ch+1)+'_gt_ids'+file_ext, dtype_str="uint32", 
                        verbose=self.cfg.TEST.VERBOSE)
                else:
                    save_tif(np.expand_dims(np.expand_dims(gt_id_img,0),-1), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS,
                        [os.path.splitext(filenames[0])[0]+'_class'+str(ch+1)+'_gt_ids.csv'], verbose=self.cfg.TEST.VERBOSE)
                
                # FP
                if len(dfs) > 0:
                    for cor in zip(fp['axis-0'].tolist(),fp['axis-1'].tolist(),fp['axis-2'].tolist()):
                        z, y, x =  cor  
                        z,y,x = int(z),int(y),int(x)
                        points_pred[z,y,x] = (0,0,255) # Blue

            # Dilate and save the predicted points for the current class 
            for i in range(points_pred.shape[0]):      
                for j in range(points_pred.shape[-1]):                                                                              
                    points_pred[i,...,j] = dilation(points_pred[i,...,j], disk(3)) 
            if file_ext in ['.hdf5', '.h5', ".zarr"]:
                write_chunked_data(points_pred, self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS, filenames[0], 
                    dtype_str="uint8", verbose=self.cfg.TEST.VERBOSE)
            else:
                save_tif(np.expand_dims(points_pred,0), self.cfg.PATHS.RESULT_DIR.DET_ASSOC_POINTS,
                    filenames, verbose=self.cfg.TEST.VERBOSE)    
                            
        return df 

    def normalize_stats(self, image_counter):
        """
        Normalize statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to average the metrics.
        """
        super().normalize_stats(image_counter)

        with open(self.cell_count_file, 'w', newline="") as file:
            csvwriter = csv.writer(file)
            csvwriter.writerow(['filename', 'cells'])
            for nr in range(len(self.cell_count_lines)):
                csvwriter.writerow([nr+1] + self.cell_count_lines[nr])
        if self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            if not self.cfg.TEST.FULL_IMG:
                self.stats['d_precision_merge_patches'] = self.stats['d_precision_merge_patches'] / image_counter
                self.stats['d_recall_merge_patches'] = self.stats['d_recall_merge_patches'] / image_counter
                self.stats['d_f1_merge_patches'] = self.stats['d_f1_merge_patches'] / image_counter
            else:
                self.stats['d_precision'] = self.stats['d_precision'] / image_counter
                self.stats['d_recall'] = self.stats['d_recall'] / image_counter
                self.stats['d_f1'] = self.stats['d_f1'] / image_counter

    def after_merge_patches(self, pred):
        """
        Steps need to be done after merging all predicted patches into the original image.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        self.detection_process(pred, self.processing_filenames, ['d_precision_merge_patches', 'd_recall_merge_patches', 'd_f1_merge_patches'])

    def after_merge_patches_by_chunks_proccess_patch(self, filename):
        """
        Place any code that needs to be done after merging all predicted patches into the original image
        but in the process made chunk by chunk. This function will operate patch by patch defined by
        ``DATA.PATCH_SIZE`` + ``DATA.PADDING``.

        Parameters
        ----------
        filename : List of str
            Filename of the predicted image H5/Zarr.
        """

        _filename, file_ext = os.path.splitext(os.path.basename(filename))
        print("Detection workflow pipeline continues for image {}".format(_filename))

        # Load H5/Zarr
        pred_file, pred = read_chunked_data(filename)

        t_dim, z_dim, c_dim, y_dim, x_dim = order_dimensions(pred.shape,
                        self.cfg.TEST.BY_CHUNKS.INPUT_IMG_AXES_ORDER)

        # Fill the new data
        z_vols = math.ceil(z_dim/self.cfg.DATA.PATCH_SIZE[0])
        y_vols = math.ceil(y_dim/self.cfg.DATA.PATCH_SIZE[1])
        x_vols = math.ceil(x_dim/self.cfg.DATA.PATCH_SIZE[2])
        total_patches = z_vols*y_vols*x_vols
        d = len(str(total_patches))
        c=1
        for z in tqdm(range(z_vols)):
            for y in range(y_vols):
                for x in range(x_vols):
                    print("Processing patch {}/{} of image".format(c, total_patches))
                    
                    print("D: z: {}-{}, y: {}-{}, x: {}-{}".format(z*self.cfg.DATA.PATCH_SIZE[0],min(z_dim,self.cfg.DATA.PATCH_SIZE[0]*(z+1)),
                        y*self.cfg.DATA.PATCH_SIZE[1],min(y_dim,self.cfg.DATA.PATCH_SIZE[1]*(y+1)),x*self.cfg.DATA.PATCH_SIZE[2],min(x_dim,self.cfg.DATA.PATCH_SIZE[2]*(x+1))))
                    
                    fname = _filename+"_patch"+str(c).zfill(d)+file_ext
                    
                    slices = [
                        slice(max(0,z*self.cfg.DATA.PATCH_SIZE[0]-self.cfg.DATA.TEST.PADDING[0]),min(z_dim,self.cfg.DATA.PATCH_SIZE[0]*(z+1)+self.cfg.DATA.TEST.PADDING[0])),
                        slice(max(0,y*self.cfg.DATA.PATCH_SIZE[1]-self.cfg.DATA.TEST.PADDING[1]),min(y_dim,self.cfg.DATA.PATCH_SIZE[1]*(y+1)+self.cfg.DATA.TEST.PADDING[1])),
                        slice(max(0,x*self.cfg.DATA.PATCH_SIZE[2]-self.cfg.DATA.TEST.PADDING[2]),min(x_dim,self.cfg.DATA.PATCH_SIZE[2]*(x+1)+self.cfg.DATA.TEST.PADDING[2])),
                        slice(None), # Channel
                    ]
                    
                    data_ordered_slices = order_dimensions(
                        slices,
                        input_order = "ZYXC",
                        output_order = self.cfg.TEST.BY_CHUNKS.INPUT_IMG_AXES_ORDER,
                        default_value = 0,
                        )

                    raw_patch = pred[data_ordered_slices]

                    current_order = np.array(range(len(pred.shape)))
                    transpose_order = order_dimensions(
                                current_order,
                                input_order= "ZYXC",
                                output_order= self.cfg.TEST.BY_CHUNKS.INPUT_IMG_AXES_ORDER,
                                default_value= np.nan)

                    transpose_order = [x for x in transpose_order if not np.isnan(x)]
                    transpose_order = np.argsort(transpose_order)
                    transpose_order = current_order[transpose_order]

                    patch = raw_patch.transpose(transpose_order)

                    df_patch = self.detection_process(patch, [fname])
                    
                    c+=1
                    
                    if df_patch is not None: # if there is at least one point detected
                        
                        if z*self.cfg.DATA.PATCH_SIZE[0]-self.cfg.DATA.TEST.PADDING[0] >=0: # if a patch was added
                            df_patch['axis-0'] = df_patch['axis-0'] - self.cfg.DATA.TEST.PADDING[0] # shift the coordinates to the correct patch position
                        if y*self.cfg.DATA.PATCH_SIZE[1]-self.cfg.DATA.TEST.PADDING[1] >=0:
                            df_patch['axis-1'] = df_patch['axis-1'] - self.cfg.DATA.TEST.PADDING[1]
                        if x*self.cfg.DATA.PATCH_SIZE[2]-self.cfg.DATA.TEST.PADDING[2] >=0:
                            df_patch['axis-2'] = df_patch['axis-2'] - self.cfg.DATA.TEST.PADDING[2]

                        df_patch = df_patch[df_patch['axis-0'] >= 0] # remove all coordinate from the previous patch
                        df_patch = df_patch[df_patch['axis-0'] < self.cfg.DATA.PATCH_SIZE[0]] # remove all coordinate from the next patch
                        df_patch = df_patch[df_patch['axis-1'] >= 0]
                        df_patch = df_patch[df_patch['axis-1'] < self.cfg.DATA.PATCH_SIZE[1]]
                        df_patch = df_patch[df_patch['axis-2'] >= 0]
                        df_patch = df_patch[df_patch['axis-2'] < self.cfg.DATA.PATCH_SIZE[2]]

                        df_patch = df_patch.reset_index(drop=True)
                        
                        # add the patch shift to the detected coordinates
                        shift = np.array([z*self.cfg.DATA.PATCH_SIZE[0], y*self.cfg.DATA.PATCH_SIZE[1], x*self.cfg.DATA.PATCH_SIZE[2]])
                        df_patch['axis-0'] = df_patch['axis-0'] + shift[0]
                        df_patch['axis-1'] = df_patch['axis-1'] + shift[1]
                        df_patch['axis-2'] = df_patch['axis-2'] + shift[2]


                        if 'df' not in locals():
                            df = df_patch.copy()
                            df['file'] = fname
                        else:
                                df_patch['file'] = fname
                                df = pd.concat([df, df_patch], ignore_index=True)

        # Apply post-processing of removing points
        if self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS and self.postpone_postproc:
            # Take point coords
            pred_coordinates = []
            coordz = df['axis-0'].tolist()
            coordy = df['axis-1'].tolist()
            coordx = df['axis-2'].tolist()
            for z,y,x in zip(coordz,coordy,coordx):
                pred_coordinates.append([z,y,x])
            radius = self.cfg.TEST.POST_PROCESSING.REMOVE_CLOSE_POINTS_RADIUS[0]
            pred_coordinates, droped_pos = remove_close_points(pred_coordinates, radius, self.cfg.DATA.TEST.RESOLUTION,
                ndim=3, return_drops=True)

            # Remove points from dataframe
            df = df.drop(droped_pos)

        # Save large csv with all point of all patches
        df = df.sort_values(by=['file'])
        df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.DET_LOCAL_MAX_COORDS_CHECK, _filename+'_all_points.csv'))

        if self.cfg.TEST.BY_CHUNKS.FORMAT == "h5":
            pred_file.close()


    def process_sample(self, norm):
        """
        Function to process a sample in the inference phase. 

        Parameters
        ----------
        norm : List of dicts
            Normalization used during training. Required to denormalize the predictions of the model.
        """
        if self.cfg.MODEL.SOURCE != "torchvision":
            super().process_sample(norm)
        else:
            # Data channel check
            if self.cfg.DATA.PATCH_SIZE[-1] != self._X.shape[-1]:
                raise ValueError("Channel of the DATA.PATCH_SIZE given {} does not correspond with the loaded image {}. "
                    "Please, check the channels of the images!".format(self.cfg.DATA.PATCH_SIZE[-1], self._X.shape[-1]))

            ##################
            ### FULL IMAGE ###
            ##################
            if self.cfg.TEST.FULL_IMG:
                # Make the prediction
                with torch.cuda.amp.autocast():
                    pred = self.model_call_func(self._X)
                del self._X 

    def torchvision_model_call(self, in_img, is_train=False):
        """
        Call a regular Pytorch model.

        Parameters
        ----------
        in_img : Tensor
            Input image to pass through the model.

        is_train : bool, optional
            Whether if the call is during training or inference. 

        Returns
        -------
        prediction : Tensor 
            Image prediction. 
        """
        filename, file_extension = os.path.splitext(self.processing_filenames[0])

        # Convert first to 0-255 range if uint16
        if in_img.dtype == torch.float32:
            if torch.max(in_img) > 1:
                in_img = (norm_range01(in_img, torch.uint8)[0]*255).to(torch.uint8)
            in_img = in_img.to(torch.uint8)
        
        # Apply TorchVision pre-processing
        in_img = self.torchvision_preprocessing(in_img)

        pred = self.model(in_img)

        bboxes = pred[0]['boxes'].cpu().numpy()
        if not is_train and len(bboxes) != 0:
            # Extract each output from prediction
            labels = pred[0]['labels'].cpu().numpy()
            scores = pred[0]['scores'].cpu().numpy()
            
            # Save all info in a csv file
            df = pd.DataFrame(zip(labels, scores, bboxes[:,0],bboxes[:,1],bboxes[:,2],bboxes[:,3]), 
                columns = ['label', 'scores', 'x1', 'y1', 'x2', 'y2'])
            df = df.sort_values(by=['label']) 
            df.to_csv(os.path.join(self.cfg.PATHS.RESULT_DIR.FULL_IMAGE, filename+".csv"), index=False)

        return None

    def after_full_image(self, pred):
        """
        Steps that must be executed after generating the prediction by supplying the entire image to the model.

        Parameters
        ----------
        pred : Torch Tensor
            Model prediction.
        """
        self.detection_process(pred, self.processing_filenames, ['d_precision', 'd_recall', 'd_f1'])

    def after_all_images(self):
        """
        Steps that must be done after predicting all images. 
        """
        super().after_all_images()

    def print_stats(self, image_counter):
        """
        Print statistics.  

        Parameters
        ----------
        image_counter : int
            Number of images to call ``normalize_stats``.
        """
        if not self.use_gt or self.cfg.MODEL.SOURCE == "torchvision": return 

        super().print_stats(image_counter)
        super().print_post_processing_stats()

        print("Detection specific metrics:")
        if self.cfg.DATA.TEST.LOAD_GT or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            if not self.cfg.TEST.FULL_IMG:
                print("Detection - Test Precision (merge patches): {}".format(self.stats['d_precision_merge_patches']))
                print("Detection - Test Recall (merge patches): {}".format(self.stats['d_recall_merge_patches']))
                print("Detection - Test F1 (merge patches): {}".format(self.stats['d_f1_merge_patches']))
            else:
                print("Detection - Test Precision (per image): {}".format(self.stats['d_precision']))
                print("Detection - Test Recall (per image): {}".format(self.stats['d_recall']))
                print("Detection - Test F1 (per image): {}".format(self.stats['d_f1']))

    def prepare_detection_data(self):
        """
        Creates detection ground truth images to train the model based on the ground truth coordinates provided.
        They will be saved in a separate folder in the root path of the ground truth. 
        """
        print("############################")
        print("#  PREPARE DETECTION DATA  #")
        print("############################")
        original_test_mask_path = None

        # Create selected channels for train data
        if self.cfg.TRAIN.ENABLE or self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.TRAIN.DETECTION_MASK_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TRAIN.DETECTION_MASK_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.TRAIN.DETECTION_MASK_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.TRAIN.DETECTION_MASK_DIR))[2]) != len(next(os.walk(self.cfg.DATA.TRAIN.GT_PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.TRAIN.GT_PATH,self.cfg.DATA.TRAIN.DETECTION_MASK_DIR))
                    create_mask = True    

            if create_mask:
                create_detection_masks(self.cfg)

        # Create selected channels for val data
        if self.cfg.TRAIN.ENABLE and not self.cfg.DATA.VAL.FROM_TRAIN:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.VAL.DETECTION_MASK_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.VAL.DETECTION_MASK_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.VAL.DETECTION_MASK_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.VAL.DETECTION_MASK_DIR))[2]) != len(next(os.walk(self.cfg.DATA.VAL.GT_PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.VAL.GT_PATH,self.cfg.DATA.VAL.DETECTION_MASK_DIR))
                    create_mask = True 
                    
            if create_mask:
                create_detection_masks(self.cfg, data_type='val')

        # Create selected channels for test data once
        if self.cfg.TEST.ENABLE and self.cfg.DATA.TEST.LOAD_GT and not self.cfg.DATA.TEST.USE_VAL_AS_TEST:
            create_mask = False
            if not os.path.isdir(self.cfg.DATA.TEST.DETECTION_MASK_DIR):
                print("You select to create detection masks from given .csv files but no file is detected in {}. "
                    "So let's prepare the data. Notice that, if you do not modify 'DATA.TEST.DETECTION_MASK_DIR' "
                    "path, this process will be done just once!".format(self.cfg.DATA.TEST.DETECTION_MASK_DIR))
                create_mask = True
            else:
                if len(next(os.walk(self.cfg.DATA.TEST.DETECTION_MASK_DIR))[2]) != len(next(os.walk(self.cfg.DATA.TEST.GT_PATH))[2]):
                    print("Different number of files found in {} and {}. Trying to create the the rest again"
                        .format(self.cfg.DATA.TEST.GT_PATH,self.cfg.DATA.TEST.DETECTION_MASK_DIR))
                    create_mask = True 
            if create_mask:
                create_detection_masks(self.cfg, data_type='test')

        opts = []
        if self.cfg.TRAIN.ENABLE:
            print("DATA.TRAIN.GT_PATH changed from {} to {}".format(self.cfg.DATA.TRAIN.GT_PATH, self.cfg.DATA.TRAIN.DETECTION_MASK_DIR))
            opts.extend(['DATA.TRAIN.GT_PATH', self.cfg.DATA.TRAIN.DETECTION_MASK_DIR])
            if not self.cfg.DATA.VAL.FROM_TRAIN:
                print("DATA.VAL.GT_PATH changed from {} to {}".format(self.cfg.DATA.VAL.GT_PATH, self.cfg.DATA.VAL.DETECTION_MASK_DIR))
                opts.extend(['DATA.VAL.GT_PATH', self.cfg.DATA.VAL.DETECTION_MASK_DIR])
        if self.cfg.TEST.ENABLE and self.cfg.DATA.TEST.LOAD_GT:
            print("DATA.TEST.GT_PATH changed from {} to {}".format(self.cfg.DATA.TEST.GT_PATH, self.cfg.DATA.TEST.DETECTION_MASK_DIR))
            opts.extend(['DATA.TEST.GT_PATH', self.cfg.DATA.TEST.DETECTION_MASK_DIR])
            original_test_mask_path = self.cfg.DATA.TEST.GT_PATH
        self.cfg.merge_from_list(opts)
        

        return original_test_mask_path


import os
import math
import numpy as np
import random
import h5py
import zarr
import matplotlib
import matplotlib.pyplot as plt
import scipy.ndimage
import copy
from PIL import Image
from tqdm import tqdm
from skimage.io import imsave, imread
from skimage import measure
from collections import namedtuple

from biapy.engine.metrics import jaccard_index_numpy, voc_calculation

def create_plots(results, metrics, job_id, chartOutDir):
    """Create loss and main metric plots with the given results.

       Parameters
       ----------
       results : Keras History object
           Record of training loss values and metrics values at successive epochs. History object is returned by Keras
           `fit() <https://keras.io/api/models/model_training_apis/#fit-method>`_ method.

       metrics : List of str
           Metrics used.

       job_id : str
           Jod identifier.

       chartOutDir : str
           Path where the charts will be stored into.

       Examples
       --------
       +-----------------------------------------------+-----------------------------------------------+
       | .. figure:: ../../img/chart_loss.png          | .. figure:: ../../img/chart_jaccard_index.png |
       |   :width: 80%                                 |   :width: 80%                                 |
       |   :align: center                              |   :align: center                              |
       |                                               |                                               |
       |   Loss values on each epoch                   |   Jaccard index values on each epoch          |
       +-----------------------------------------------+-----------------------------------------------+
    """

    print("Creating training plots . . .")
    os.makedirs(chartOutDir, exist_ok=True)

    # For matplotlib errors in display
    os.environ['QT_QPA_PLATFORM']='offscreen'

    # Loss
    plt.plot(results['loss'])
    if 'val_loss' in results:
        plt.plot(results['val_loss'])
    plt.title('Model JOBID=' + job_id + ' loss')
    plt.ylabel('Value')
    plt.xlabel('Epoch')
    if 'val_loss' in results:
        plt.legend(['Train loss', 'Val. loss'], loc='upper left')
    else:
        plt.legend(['Train loss'], loc='upper left')
    plt.savefig(os.path.join(chartOutDir, job_id + '_loss.png'))
    plt.clf()

    # Metric
    for i in range(len(metrics)):
        plt.plot(results[metrics[i]])
        plt.plot(results['val_' + metrics[i]])
        plt.title('Model JOBID=' + job_id + " " + metrics[i])
        plt.ylabel('Value')
        plt.xlabel('Epoch')
        plt.legend([f'Train {metrics[i]}', f'Val. {metrics[i]}'], loc='upper left')
        plt.savefig(os.path.join(chartOutDir, job_id + '_' + metrics[i] +'.png'))
        plt.clf()


def threshold_plots(preds_test, Y_test, n_dig, job_id, job_file, char_dir, r_val=0.5):
    """Create a plot with the different metric values binarizing the prediction with different thresholds, from ``0.1``
       to ``0.9``.

       Parameters
       ----------
       preds_test : 4D Numpy array
           Predictions made by the model. E.g. ``(num_of_images, y, x, channels)``.

       Y_test : 4D Numpy array
           Ground truth of the data. E.g. ``(num_of_images, y, x, channels)``.

       n_dig : int
           The number of digits used for encoding temporal indices (e.g. ``3``). Used by the DET calculation binary.

       job_id : str
           Id of the job.

       job_file : str
           Id and run number of the job.

       char_dir : str
           Path to store the charts generated.

       r_val : float, optional
           Threshold values to return.

       Returns
       -------
       t_jac : float
           Value of the Jaccard index when the threshold is ``r_val``.

       t_voc : float
           Value of VOC when the threshold is ``r_val``.

       Examples
       --------
       ::

           jac, voc = threshold_plots(
               preds_test, Y_test, det_eval_ge_path, det_eval_path, det_bin,
               n_dig, args.job_id, '278_3', char_dir)

       Will generate 3 charts, one per each metric: IoU and VOC. In the x axis represents the 9 different
       thresholds applied, that is: ``0.1, 0.2, 0.3, ..., 0.9``. The y axis is the value of the metric in each chart. For
       instance, the Jaccard/IoU chart will look like this:

       .. image:: ../../img/278_3_threshold_Jaccard.png
           :width: 60%
           :align: center

       In this example, the best value, ``0.868``, is obtained with a threshold of ``0.4``.
    """

    char_dir = os.path.join(char_dir, "t_" + job_file)

    t_jac = np.zeros(9)
    t_voc = np.zeros(9)
    objects = []
    r_val_pos = 0

    for i, t in enumerate(np.arange(0.1,1.0,0.1)):

        if t == r_val:
            r_val_pos = i

        objects.append(str('%.2f' % float(t)))

        # Threshold images
        bin_preds_test = (preds_test > t).astype(np.uint8)

        # Metrics (Jaccard + VOC)
        print("Calculate metrics . . .")
        t_jac[i] = jaccard_index_numpy(Y_test, bin_preds_test)
        t_voc[i] = voc_calculation(Y_test, bin_preds_test, t_jac[i])

        print("t_jac[{}]: {}".format(i, t_jac[i]))
        print("t_voc[{}]: {}".format(i, t_voc[i]))

    # For matplotlib errors in display
    os.environ['QT_QPA_PLATFORM']='offscreen'

    os.makedirs(char_dir, exist_ok=True)

    # Plot Jaccard values
    plt.clf()
    plt.plot(objects, t_jac)
    plt.title('Model JOBID=' + job_file + ' Jaccard', y=1.08)
    plt.ylabel('Value')
    plt.xlabel('Threshold')
    for k, point in enumerate(zip(objects, t_jac)):
        plt.text(point[0], point[1], '%.3f' % float(t_jac[k]))
    plt.savefig(os.path.join(char_dir, job_file + '_threshold_Jaccard.png'))
    plt.clf()

    # Plot VOC values
    plt.plot(objects, t_voc)
    plt.title('Model JOBID=' + job_file + ' VOC', y=1.08)
    plt.ylabel('Value')
    plt.xlabel('Threshold')
    for k, point in enumerate(zip(objects, t_voc)):
        plt.text(point[0], point[1], '%.3f' % float(t_voc[k]))
    plt.savefig(os.path.join(char_dir, job_file + '_threshold_VOC.png'))
    plt.clf()

    return  t_jac[r_val_pos], t_voc[r_val_pos]


def save_tif(X, data_dir=None, filenames=None, verbose=True):
    """Save images in the given directory.

       Parameters
       ----------
       X : 4D/5D numpy array
           Data to save as images. The first dimension must be the number of images. E.g.
           ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.

       data_dir : str, optional
           Path to store X images.

       filenames : List, optional
           Filenames that should be used when saving each image.

       verbose : bool, optional
            To print saving information.
    """

    if verbose:
        s = X.shape if not isinstance(X, list) else X[0].shape
        print("Saving {} data as .tif in folder: {}".format(s, data_dir))

    os.makedirs(data_dir, exist_ok=True)
    if filenames is not None:
        if len(filenames) != len(X):
            raise ValueError("Filenames array and length of X have different shapes: {} vs {}".format(len(filenames),len(X)))

    if not isinstance(X, list):
        _dtype = X.dtype if X.dtype in [np.uint8, np.uint16, np.float32] else np.float32
        ndims = X.ndim
    else:
        _dtype = X[0].dtype if X[0].dtype in [np.uint8, np.uint16, np.float32] else np.float32
        ndims = X[0].ndim

    d = len(str(len(X)))
    for i in tqdm(range(len(X)), leave=False):
        if filenames is None:
            f = os.path.join(data_dir, str(i).zfill(d)+'.tif')
        else:
            f = os.path.join(data_dir, os.path.splitext(filenames[i])[0]+'.tif')
        if ndims == 4:
            if not isinstance(X, list):
                aux = np.expand_dims(np.expand_dims(X[i],0).transpose((0,3,1,2)), -1).astype(_dtype)
            else:
                aux = np.expand_dims(np.expand_dims(X[i][0],0).transpose((0,3,1,2)), -1).astype(_dtype)
        else:
            if not isinstance(X, list):
                aux = np.expand_dims(X[i].transpose((0,3,1,2)), -1).astype(_dtype)
            else:
                aux = np.expand_dims(X[i][0].transpose((0,3,1,2)), -1).astype(_dtype)
        try:
            imsave(f, np.expand_dims(aux, 0), imagej=True, metadata={'axes': 'TZCYXS'}, check_contrast=False, compression=('zlib', 1))
        except:
            imsave(f, np.expand_dims(aux, 0), imagej=True, metadata={'axes': 'TZCYXS'}, check_contrast=False)


def save_tif_pair_discard(X, Y, data_dir=None, suffix="", filenames=None, discard=True, verbose=True):
    """Save images in the given directory.

       Parameters
       ----------
       X : 4D/5D numpy array
           Data to save as images. The first dimension must be the number of images. E.g.
           ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.

       Y : 4D/5D numpy array
           Data mask to save. The first dimension must be the number of images. E.g.
           ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.

       data_dir : str, optional
           Path to store X images.

       suffix : str, optional
           Suffix to apply on output directory.

       filenames : List, optional
           Filenames that should be used when saving each image.

       discard : bool, optional
           Whether to discard image/mask pairs if the mask has no label information.

       verbose : bool, optional
            To print saving information.
    """

    if verbose:
        s = X.shape if not isinstance(X, list) else X[0].shape
        print("Saving {} data as .tif in folder: {}".format(s, data_dir))

    os.makedirs(os.path.join(data_dir, 'x'+suffix), exist_ok=True)
    os.makedirs(os.path.join(data_dir, 'y'+suffix), exist_ok=True)
    if filenames is not None:
        if len(filenames) != len(X):
            raise ValueError("Filenames array and length of X have different shapes: {} vs {}".format(len(filenames),len(X)))

    _dtype = X.dtype if X.dtype in [np.uint8, np.uint16, np.float32] else np.float32
    d = len(str(len(X)))
    for i in tqdm(range(X.shape[0]), leave=False):
        if len(np.unique(Y[i])) >= 2 or not discard:
            if filenames is None:
                f1 = os.path.join(data_dir, 'x'+suffix, str(i).zfill(d)+'.tif')
                f2 = os.path.join(data_dir, 'y'+suffix, str(i).zfill(d)+'.tif')
            else:
                f1 = os.path.join(data_dir, 'x'+suffix, os.path.splitext(filenames[i])[0]+'.tif')
                f2 = os.path.join(data_dir, 'y'+suffix, os.path.splitext(filenames[i])[0]+'.tif')
            if X.ndim == 4:
                aux = np.expand_dims(np.expand_dims(X[i],0).transpose((0,3,1,2)), -1).astype(_dtype)
            else:
                aux = np.expand_dims(X[i].transpose((0,3,1,2)), -1).astype(_dtype)
            imsave(f1, np.expand_dims(aux, 0), imagej=True, metadata={'axes': 'TZCYXS'}, check_contrast=False, compression=('zlib', 1))
            if Y.ndim == 4:
                aux = np.expand_dims(np.expand_dims(Y[i],0).transpose((0,3,1,2)), -1).astype(_dtype)
            else:
                aux = np.expand_dims(Y[i].transpose((0,3,1,2)), -1).astype(_dtype)
            imsave(f2, np.expand_dims(aux, 0), imagej=True, metadata={'axes': 'TZCYXS'}, check_contrast=False, compression=('zlib', 1))


def save_img(X=None, data_dir=None, Y=None, mask_dir=None, scale_mask=True,
             prefix="", extension=".png", filenames=None):
    """Save images in the given directory.

       Parameters
       ----------
       X : 4D numpy array, optional
           Data to save as images. The first dimension must be the number of images. E.g. ``(num_of_images, y, x, channels)``.

       data_dir : str, optional
           Path to store X images.

       Y : 4D numpy array, optional
           Masks to save as images. The first dimension must be the number of images. E.g. ``(num_of_images, y, x, channels)``.

       scale_mask : bool, optional
           To allow mask be multiplied by 255.

       mask_dir : str, optional
           Path to store Y images.

       prefix : str, optional
           Path to store generated charts.

       filenames : list, optional
           Filenames that should be used when saving each image. If any provided each image should be named as:
           ``prefix + "_x_" + image_number + extension`` when ``X.ndim < 4`` and ``prefix + "_x_" + image_number +
           "_" + slice_numger + extension`` otherwise. E.g. ``prefix_x_000.png`` when ``X.ndim < 4`` or
           ``prefix_x_000_000.png`` when ``X.ndim >= 4``.  The same applies to ``Y``.
    """

    if prefix == "":
        p_x = "x_"
        p_y = "y_"
    else:
        p_x = prefix + "_x_"
        p_y = prefix + "_y_"

    if X is not None:
        if data_dir is not None:
            os.makedirs(data_dir, exist_ok=True)
        else:
            print("Not data_dir provided so no image will be saved!")
            return

        print("Saving images in {}".format(data_dir))

        v = 1 if np.max(X) > 2 else 255
        if X.ndim > 4:
            d = len(str(X.shape[0]*X.shape[3]))
            for i in tqdm(range(X.shape[0])):
                for j in range(X.shape[3]):
                    if X.shape[-1] == 1:
                        im = Image.fromarray((X[i,:,:,j,0]*v).astype(np.uint8))
                        im = im.convert('L')
                    else:
                        im = Image.fromarray((X[i,:,:,j]*v).astype(np.uint8), 'RGB')

                    if filenames is None:
                        f = os.path.join(data_dir, p_x + str(i).zfill(d) + "_" + str(j).zfill(d) + extension)
                    else:
                        f = os.path.join(data_dir, filenames[(i*j)+j] + extension)
                    im.save(f)
        else:
            d = len(str(X.shape[0]))
            for i in tqdm(range(X.shape[0])):
                if X.shape[-1] == 1:
                    im = Image.fromarray((X[i,:,:,0]*v).astype(np.uint8))
                    im = im.convert('L')
                else:
                    im = Image.fromarray((X[i]*v).astype(np.uint8), 'RGB')

                if filenames is None:
                    f = os.path.join(data_dir, p_x + str(i).zfill(d) + extension)
                else:
                    f = os.path.join(data_dir, filenames[i] + extension)
                im.save(f)

    if Y is not None:
        if mask_dir is not None:
            os.makedirs(mask_dir, exist_ok=True)
        else:
            print("Not mask_dir provided so no image will be saved!")
            return

        print("Saving images in {}".format(mask_dir))

        v = 1 if np.max(Y) > 2 or not scale_mask else 255
        if Y.ndim > 4:
            d = len(str(Y.shape[0]*Y.shape[3]))
            for i in tqdm(range(Y.shape[0])):
                for j in range(Y.shape[3]):
                    for k in range(Y.shape[-1]):
                        im = Image.fromarray((Y[i,:,:,j,k]*v).astype(np.uint8))
                        im = im.convert('L')
                        if filenames is None:
                            c = "" if Y.shape[-1] == 1 else "_c"+str(j)
                            f = os.path.join(mask_dir, p_y + str(i).zfill(d) + "_" + str(j).zfill(d)+c+extension)
                        else:
                            f = os.path.join(data_dir, filenames[(i*j)+j] + extension)

                        im.save(f)
        else:
            d = len(str(Y.shape[0]))
            for i in tqdm(range(0, Y.shape[0])):
                for j in range(Y.shape[-1]):
                    im = Image.fromarray((Y[i,:,:,j]*v).astype(np.uint8))
                    im = im.convert('L')

                    if filenames is None:
                        c = "" if Y.shape[-1] == 1 else "_c"+str(j)
                        f = os.path.join(mask_dir, p_y+str(i).zfill(d)+c+extension)
                    else:
                        f = os.path.join(mask_dir, filenames[i] + extension)

                    im.save(f)


def make_weight_map(label, binary = True, w0 = 10, sigma = 5):
    """Generates a weight map in order to make the U-Net learn better the borders of cells and distinguish individual
       cells that are tightly packed. These weight maps follow the methodology of the original U-Net paper.

       Based on `unet/py_files/helpers.py <https://github.com/deepimagej/python4deepimagej/blob/499955a264e1b66c4ed2c014cb139289be0e98a4/unet/py_files/helpers.py>`_.

       Parameters
       ----------

       label : 3D numpy array
          Corresponds to a label image. E.g. ``(y, x, channels)``.

       binary : bool, optional
          Corresponds to whether or not the labels are binary.

       w0 : float, optional
          Controls for the importance of separating tightly associated entities.

       sigma : int, optional
          Represents the standard deviation of the Gaussian used for the weight map.

       Examples
       --------

       Notice that weight has been defined where the objects are almost touching
       each other.

       .. image:: ../../img/weight_map.png
           :width: 650
           :align: center
    """

    # Initialization.
    lab = np.array(label)
    lab_multi = lab

    if len(lab.shape) == 3:
        lab = lab[:, :, 0]

    # Get shape of label.
    rows, cols = lab.shape

    if binary:

        # Converts the label into a binary image with background = 0
        # and cells = 1.
        lab[lab == 255] = 1

        # Builds w_c which is the class balancing map. In our case, we want
        # cells to have weight 2 as they are more important than background
        # which is assigned weight 1.
        w_c = np.array(lab, dtype=float)
        w_c[w_c == 1] = 1
        w_c[w_c == 0] = 0.5

        # Converts the labels to have one class per object (cell).
        lab_multi = measure.label(lab, neighbors = 8, background = 0)
    else:

        # Converts the label into a binary image with background = 0.
        # and cells = 1.
        lab[lab > 0] = 1

        # Builds w_c which is the class balancing map. In our case, we want
        # cells to have weight 2 as they are more important than background
        # which is assigned weight 1.
        w_c = np.array(lab, dtype=float)
        w_c[w_c == 1] = 1
        w_c[w_c == 0] = 0.5
    components = np.unique(lab_multi)

    n_comp = len(components)-1

    maps = np.zeros((n_comp, rows, cols))

    map_weight = np.zeros((rows, cols))

    if n_comp >= 2:
        for i in range(n_comp):

            # Only keeps current object.
            tmp = (lab_multi == components[i+1])

            # Invert tmp so that it can have the correct distance.
            # transform
            tmp = ~tmp

            # For each pixel, computes the distance transform to
            # each object.
            maps[i][:][:] = scipy.ndimage.distance_transform_edt(tmp)

        maps = np.sort(maps, axis=0)

        # Get distance to the closest object (d1) and the distance to the second
        # object (d2).
        d1 = maps[0][:][:]
        d2 = maps[1][:][:]

        map_weight = w0*np.exp(-((d1+d2)**2)/(2*(sigma**2)) ) * (lab==0).astype(int);

    map_weight += w_c

    return map_weight


def do_save_wm(labels, path, binary = True, w0 = 10, sigma = 5):
    """Retrieves the label images, applies the weight-map algorithm and save the weight maps in a folder. Uses
       internally :meth:`util.make_weight_map`.

       Based on `deepimagejunet/py_files/helpers.py <https://github.com/deepimagej/python4deepimagej/blob/499955a264e1b66c4ed2c014cb139289be0e98a4/unet/py_files/helpers.py>`_.

       Parameters
       ----------
       labels : 4D numpy array
           Corresponds to given label images. E.g. ``(num_of_images, y, x, channels)``.

       path : str
           Refers to the path where the weight maps should be saved.

       binary : bool, optional
           Corresponds to whether or not the labels are binary.

       w0 : float, optional
           Controls for the importance of separating tightly associated entities.

       sigma : int, optional
           Represents the standard deviation of the Gaussian used for the weight
           map.
    """

    # Copy labels.
    labels_ = copy.deepcopy(labels)

    # Perform weight maps.
    for i in range(len(labels_)):
        labels_[i] = make_weight_map(labels[i].copy(), binary, w0, sigma)

    maps = np.array(labels_)

    n, rows, cols = maps.shape

    # Resize correctly the maps so that it can be used in the model.
    maps = maps.reshape((n, rows, cols, 1))

    # Count number of digits in n. This is important for the number
    # of leading zeros in the name of the maps.
    n_digits = len(str(n))

    # Save path with correct leading zeros.
    path_to_save = path + "weight/{b:0" + str(n_digits) + "d}.npy"

    # Saving files as .npy files.
    for i in range(len(labels_)):
        np.save(path_to_save.format(b=i), labels_[i])

    return None


def foreground_percentage(mask, class_tag):
    """Percentage of pixels that corresponds to the class in the given image.

       Parameters
       ----------
       mask : 2D Numpy array
           Image mask to analize.

       class_tag : int
           Class to find in the image.

       Returns
       -------
       x : float
           Percentage of pixels that corresponds to the class. Value between ``0``
           and ``1``.
    """

    c = 0
    for i in range(0, mask.shape[0]):
        for j in range(0, mask.shape[1]):
            if mask[i, j, 0] == class_tag:
                c = c + 1

    return c/(mask.shape[0]*mask.shape[1])


def divide_images_on_classes(data, data_mask, out_dir, num_classes=2, th=0.8):
    """Create a folder for each class where the images that have more pixels labeled as the class (in percentage) than
       the given threshold will be stored.

       Parameters
       ----------
       data : 4D numpy array
           Data to save as images. The first dimension must be the number of images. E. g.``(num_of_images, y, x, channels)``.

       data_mask : 4D numpy array
           Data mask to save as images.  The first dimension must be the number of images. E. g. ``(num_of_images, y, x, channels)``.

       out_dir : str
           Path to save the images.

       num_classes : int, optional
           Number of classes.

       th : float, optional
           Percentage of the pixels that must be labeled as a class to save it inside that class folder.
    """

    # Create the directories
    for i in range(num_classes):
        os.makedirs(os.path.join(out_dir, "x", "class"+str(i)), exist_ok=True)
        os.makedirs(os.path.join(out_dir, "y", "class"+str(i)), exist_ok=True)

    print("Dividing provided data into {} classes . . .".format(num_classes))
    d = len(str(data.shape[0]))
    for i in tqdm(range(data.shape[0])):
        # Assign the image to a class if it has, in percentage, more pixels of
        # that class than the given threshold
        for j in range(num_classes):
            t = foreground_percentage(data_mask[i], j)
            if t > th:
                im = Image.fromarray(data[i,:,:,0])
                im = im.convert('L')
                im.save(os.path.join(os.path.join(out_dir, "x", "class"+str(j)), "im_" + str(i).zfill(d) + ".png"))
                im = Image.fromarray(data_mask[i,:,:,0]*255)
                im = im.convert('L')
                im.save(os.path.join(os.path.join(out_dir, "y", "class"+str(j)), "mask_" + str(i).zfill(d) + ".png"))


def save_filters_of_convlayer(model, out_dir, l_num=None, name=None, prefix="", img_per_row=8):
    """Create an image of the filters learned by a convolutional layer. One can identify the layer with ``l_num`` or
       ``name`` args. If both are passed ``name`` will be prioritized.

       Inspired by https://machinelearningmastery.com/how-to-visualize-filters-and-feature-maps-in-convolutional-neural-networks

       Parameters
       ----------
       model : Keras Model
           Model where the layers are stored.

       out_dir : str
           Path where the image will be stored.

       l_num : int, optional
           Number of the layer to extract filters from.

       name : str, optional
           Name of the layer to extract filters from.

       prefix : str, optional
           Prefix to add to the output image name.

       img_per_row : int, optional
           Filters per row on the image.

       Raises
       ------
       ValueError
           if ``l_num`` and ``name`` not provided.

       Examples
       --------
       To save the filters learned by the layer called ``conv1`` one can call
       the function as follows ::

           save_filters_of_convlayer(model, char_dir, name="conv1", prefix="model")

       That will save in ``out_dir`` an image like this:

       .. image:: ../../img/save_filters.png
           :width: 60%
           :align: center
    """

    if l_num is None and name is None:
        raise ValueError("One between 'l_num' or 'name' must be provided")

    # For matplotlib errors in display
    os.environ['QT_QPA_PLATFORM']='offscreen'

    # Find layer number of the layer named by 'name' variable
    if name is not None:
        pos = 0
        for layer in model.layers:
            if name == layer.name:
                break
            pos += 1
        l_num = pos

    filters, biases = model.layers[l_num].get_weights()

    # normalize filter values to 0-1 so we can visualize them
    f_min, f_max = filters.min(), filters.max()
    filters = (filters - f_min) / (f_max - f_min)

    rows = int(math.floor(filters.shape[3]/img_per_row))
    i = 0
    for r in range(rows):
        for c in range(img_per_row):
            ax = plt.subplot(rows, img_per_row, i+1)
            ax.set_xticks([])
            ax.set_yticks([])
            f = filters[:,:,0,i]
            plt.imshow(filters[:,:,0,i], cmap='gray')

            i += 1

    prefix += "_" if prefix != "" else prefix
    plt.savefig(os.path.join(out_dir, prefix + 'f_layer' + str(l_num) + '.png'))
    plt.clf()


def check_masks(path, n_classes=2):
    """Check Whether the data masks have the correct labels inspection a few random images of the given path. If the
       function gives no error one should assume that the masks are correct.

       Parameters
       ----------
       path : str
           Path to the data mask.

       n_classes : int, optional
           Maximum classes that the masks must contain.
    """

    print("Checking ground truth classes in {} . . .".format(path))

    ids = sorted(next(os.walk(path))[2])

    # Check only 4 random images or less if there are not as many
    num_sample = [4, len(ids)]
    numbers = random.sample(range(0, len(ids)), min(num_sample))
    for i in numbers:
        img = imread(os.path.join(path, ids[i]))
        values, _ = np.unique(img, return_counts=True)
        if len(values) > n_classes :
            raise ValueError("Error: given mask ({}) has more classes than specified in 'MODEL.N_CLASSES'."
                             "That variable value need to be counting with the background class." 
                             " E.g. if mask has [0,1,2] values 'MODEL.N_CLASSES' should be 3.\n"
                             "Values found: {}".format(os.path.join(path, ids[i]), values))

def img_to_onehot_encoding(img, num_classes=2):
    """Converts image given into one-hot encode format.

       The opposite function is :func:`~onehot_encoding_to_img`.

       Parameters
       ----------
       img : Numpy 3D/4D array
           Image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       num_classes : int, optional
           Number of classes to distinguish.

       Returns
       -------
       one_hot_labels : Numpy 3D/4D array
           Data one-hot encoded. E.g. ``(y, x, num_classes)`` or ``(z, y, x, num_classes)``.
    """

    if img.ndim == 4:
        shape = img.shape[:3]+(num_classes,)
    else:
        shape = img.shape[:2]+(num_classes,)

    encoded_image = np.zeros(shape, dtype=np.int8)

    for i in range(num_classes):
        if img.ndim == 4:
            encoded_image[:,:,:,i] = np.all(img.reshape((-1,1)) == i, axis=1).reshape(shape[:3])
        else:
            encoded_image[:,:,i] = np.all(img.reshape((-1,1)) == i, axis=1).reshape(shape[:2])

    return encoded_image


def onehot_encoding_to_img(encoded_image):
    """Converts one-hot encode image into an image with jus tone channel and all the classes represented by an integer.

       The opposite function is :func:`~img_to_onehot_encoding`.

       Parameters
       ----------
       encoded_image : Numpy 3D/4D array
           Image. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       Returns
       -------
       img : Numpy 3D/4D array
           Data one-hot encoded. E.g. ``(z, y, x, num_classes)``.
    """

    if encoded_image.ndim == 4:
        shape = encoded_image.shape[:3]+(1,)
    else:
        shape = encoded_image.shape[:2]+(1,)

    img = np.zeros(shape, dtype=np.int8)
    for i in range(img.shape[-1]):
        img[encoded_image[...,i] == 1] = i

    return img


def load_data_from_dir(data_dir, crop=False, crop_shape=None, overlap=(0,0), padding=(0,0), return_filenames=False,
                       reflect_to_complete_shape=False, check_channel=True, convert_to_rgb=False, check_drange=True,
                       preprocess_cfg=None, is_mask=False, preprocess_f=None):
    """Load data from a directory. If ``crop=False`` all the data is suposed to have the same shape.

    Parameters
    ----------
    data_dir : str
        Path to read the data from.

    crop : bool, optional
        Crop each image into desired shape pointed by ``crop_shape``.

    crop_shape : Tuple of 3 ints, optional
        Shape of the crop to be made. E.g. ``(y, x, channels)``.

    overlap : Tuple of 2 floats, optional
        Amount of minimum overlap on x and y dimensions. The values must  be on range ``[0, 1)``, that is, ``0%`` or
        ``99%`` of overlap. E. g. ``(y, x)``.

    padding : Tuple of 2 ints, optional
        Size of padding to be added on each axis ``(y, x)``. E.g. ``(24, 24)``.

    return_filenames : bool, optional
        Return a list with the loaded filenames. Useful when you need to save them afterwards with the same names as
        the original ones.

    reflect_to_complete_shape : bool, optional
        Whether to increase the shape of the dimension that have less size than selected patch size padding it with
        'reflect'.

    check_channel : bool, optional
        Whether to check if the crop_shape channel matches with the loaded images' one. 
        
    convert_to_rgb : bool, optional
        In case RGB images are expected, e.g. if ``crop_shape`` channel is 3, those images that are grayscale are 
        converted into RGB.
           
    check_drange : bool, optional
        Whether to check if the data loaded is in the same range. 
    
    preprocess_cfg : dict, optional
        Configuration parameters for preprocessing, is necessary in case you want to apply any preprocessing.
    
    is_mask : bool, optional
        Whether the data are masks. It is used to control the preprocessing of the data.
    
    preprocess_f : function, optional
        The preprocessing function, is necessary in case you want to apply any preprocessing.

    Returns
    -------
    data : 4D Numpy array or list of 3D Numpy arrays
        Data loaded. E.g. ``(num_of_images, y, x, channels)`` if all files have same shape, otherwise a list of
        ``(y, x, channels)`` arrays will be returned.

    data_shape : List of tuples
        Shapes of all 3D images readed. Useful to reconstruct the original images together with ``crop_shape``.

    crop_shape : List of tuples
        Shape of the loaded 3D images after cropping. Useful to reconstruct the original images together with
        ``data_shape``.

    filenames : List of str, optional
        Loaded filenames.

    Examples
    --------
    ::

        # EXAMPLE 1
        # Case where we need to load 165 images of shape (1024, 768)
        data_path = "data/train/x"

        load_data_from_dir(data_path)
        # The function will print the shape of the created array. In this example:
        #     *** Loaded data shape is (165, 768, 1024, 1)
        # Notice height and width swap because of Numpy ndarray terminology


        # EXAMPLE 2
        # Case where we need to load 165 images of shape (1024, 768) but
        # cropping them into (256, 256, 1) patches
        data_path = "data/train/x"
        crop_shape = (256, 256, 1)

        load_data_from_dir(data_path, crop=True, crop_shape=crop_shape)
        # The function will print the shape of the created array. In this example:
        #     *** Loaded data shape is (1980, 256, 256, 1)
    """

    if preprocess_f != None and preprocess_cfg == None:
        raise ValueError("The preprocessing configuration ('preprocess_cfg') is missing.")

    if crop:
        from biapy.data.data_2D_manipulation import crop_data_with_overlap

    print("Loading data from {}".format(data_dir))
    ids = sorted(next(os.walk(data_dir))[2])
    data = []
    data_shape = []
    c_shape = []
    filenames = []

    if len(ids) == 0:
        raise ValueError("No images found in dir {}".format(data_dir))

    for n, id_ in tqdm(enumerate(ids), total=len(ids)):
        if id_.endswith('.npy'):
            img = np.load(os.path.join(data_dir, id_))
        elif id_.endswith('.hdf5') or id_.endswith('.h5'):
            img = h5py.File(os.path.join(data_dir, id_),'r')
            img = np.array(img[list(img)[0]])
        else:
            img = imread(os.path.join(data_dir, id_))
        img = np.squeeze(img)

        if img.ndim > 3:
            raise ValueError("Read image seems to be 3D: {}. Path: {}".format(img.shape, os.path.join(data_dir, id_)))

        filenames.append(id_)

        if img.ndim == 2:
            img = np.expand_dims(img, -1)
        else:
            if img.shape[0] <= 3: img = img.transpose((1,2,0))  

        if reflect_to_complete_shape: img = pad_and_reflect(img, crop_shape, verbose=False)

        if crop_shape is not None and check_channel:
            if crop_shape[-1] != img.shape[-1]:
                if crop_shape[-1] == 3 and convert_to_rgb:
                    img = np.repeat(img, 3, axis=-1)
                else:
                    raise ValueError("Channel of the patch size given {} does not correspond with the loaded image {}. "
                        "Please, check the channels of the images!".format(crop_shape[-1], img.shape[-1]))

        if preprocess_f == None:
            data_shape.append(img.shape)
            img = np.expand_dims(img, axis=0)
            if crop and img[0].shape != crop_shape[:2]+(img.shape[-1],):
                img = crop_data_with_overlap(img, crop_shape[:2]+(img.shape[-1],), overlap=overlap, padding=padding,
                                            verbose=False)
            c_shape.append(img.shape)
        data.append(img)
        
    if preprocess_f != None:
        if is_mask:
            # data contains masks
            data = preprocess_f(preprocess_cfg, y_data = data, is_2d = True, is_y_mask = is_mask)
        else:
            data = preprocess_f(preprocess_cfg, x_data = data, is_2d = True)
    
        _data = []
        for img_num in range(len(data)):
            img = data[img_num]
            data_shape.append(img.shape)
            img = np.expand_dims(img, axis=0)
            if crop and img[0].shape != crop_shape[:2]+(img.shape[-1],):
                img = crop_data_with_overlap(img, crop_shape[:2]+(img.shape[-1],), overlap=overlap, padding=padding,
                                            verbose=False)
            c_shape.append(img.shape)
            _data.append(img)
        del data
        data = _data

    same_shape = True
    s = data[0].shape
    dtype = data[0].dtype
    drange = data_range(data[0])
    for i in range(1,len(data)):
        if check_drange and drange != data_range(data[i]):
            raise ValueError("Input images ({} vs {}) seem to have different data ranges ({} and {} found) Please check it "
                "and ensure all images have same data type"
                .format(filenames[0], filenames[i], drange, data_range(data[i])))
        if s != data[i].shape:
            same_shape = False

    if crop or same_shape:
        data = np.concatenate(data)
        print("*** Loaded data shape is {}".format(data.shape))
    else:
        print("Not all samples seem to have the same shape. Number of samples: {}".format(len(data)))
        print("*** First sample shape is {}".format(data[0].shape[1:]))

    if return_filenames:
        return data, data_shape, c_shape, filenames
    else:
        return data, data_shape, c_shape


def load_ct_data_from_dir(data_dir, shape=None):
    """Load CT data from a directory.

       Parameters
       ----------
       data_dir : str
           Path to read the data from.

       shape : 3D int tuple, optional
           Shape of the data to load. If is not provided the shape is calculated automatically looping over all data
           files and it will be  the maximum value found per axis. So, given the value the process should be faster.
           E.g. ``(y, x, channels)``.

       Returns
       -------
       data : 4D Numpy array
           Data loaded. E.g. ``(num_of_images, y, x, channels)``.

       Examples
       --------
       ::

           # EXAMPLE 1
           # Case where we need to load 165 images of shape (1024, 768)
           data_path = "data/train/x"
           data_shape = (1024, 768, 1)

           load_data_from_dir(data_path, data_shape)

           # The function will print list's first position array's shape. In this example:
           #     *** Loaded data[0] shape is (165, 768, 1024, 1)
           # Notice height and width swap because of Numpy ndarray terminology
    """
    import nibabel as nib

    print("Loading data from {}".format(data_dir))
    ids = sorted(next(os.walk(data_dir))[2])

    if shape is None:
        # Determine max in each dimension first
        max_x = 0
        max_y = 0
        max_z = 0
        for n, id_ in tqdm(enumerate(ids), total=len(ids)):
            img = nib.load(os.path.join(data_dir, id_))
            img = np.array(img.dataobj)

            max_x = img.shape[0] if max_x < img.shape[0] else max_x
            max_y = img.shape[1] if max_y < img.shape[1] else max_y
            max_z = img.shape[2] if max_z < img.shape[2] else max_z
        _shape = (max_x, max_y, max_z)
    else:
        _shape = shape

    # Create the array
    data = np.zeros((len(ids), ) + _shape, dtype=np.float32)
    for n, id_ in tqdm(enumerate(ids), total=len(ids)):
        img = nib.load(os.path.join(data_dir, id_))
        img = np.array(img.dataobj)
        data[n,0:img.shape[0],0:img.shape[1],0:img.shape[2]] = img

    print("*** Loaded data shape is {}".format(data.shape))
    return data


def load_3d_images_from_dir(data_dir, crop=False, crop_shape=None, verbose=False, overlap=(0,0,0), padding=(0,0,0),
        median_padding=False, reflect_to_complete_shape=False, check_channel=True, convert_to_rgb=False, check_drange=True,
        return_filenames=False, preprocess_cfg=None, is_mask=False, preprocess_f=None):
    """Load data from a directory.

    Parameters
    ----------
    data_dir : str
        Path to read the data from.

    crop : bool, optional
        Crop each 3D image when readed.

    crop_shape : Tuple of 4 ints, optional
        Shape of the subvolumes to create when cropping.  E.g. ``(z, y, x, channels)``.

    verbose : bool, optional
        Whether to enable verbosity.

    overlap : Tuple of 3 floats, optional
        Amount of minimum overlap on z, y and x dimensions. The values must be on range ``[0, 1)``, that is, ``0%``
        or ``99%`` of overlap. E.g. ``(z, y, x)``.

    padding : Tuple of 3 ints, optional
        Size of padding to be added on each axis ``(z, y, x)``. E.g. ``(24, 24, 24)``.

    median_padding : bool, optional
        If ``True`` the padding value is the median value. If ``False``, the added values are zeroes.

    reflect_to_complete_shape : bool, optional
        Whether to increase the shape of the dimension that have less size than selected patch size padding it with
        'reflect'.

    check_channel : bool, optional
        Whether to check if the crop_shape channel matches with the loaded images' one.

    convert_to_rgb : bool, optional
        In case RGB images are expected, e.g. if ``crop_shape`` channel is 3, those images that are grayscale are 
        converted into RGB.

    check_drange : bool, optional
        Whether to check if the data loaded is in the same range. 

    preprocess_cfg : dict, optional
        Configuration parameters for preprocessing, is necessary in case you want to apply any preprocessing.
    
    is_mask : bool, optional
        Whether the data are masks. It is used to control the preprocessing of the data.
    
    preprocess_f : function, optional
        The preprocessing function, is necessary in case you want to apply any preprocessing.

    return_filenames : bool, optional
        Return a list with the loaded filenames. Useful when you need to save them afterwards with the same names as
        the original ones.

    Returns
    -------
    data : 5D Numpy array or list of 4D Numpy arrays
        Data loaded. E.g. ``(num_of_images, z, y, x, channels)`` if all files have same shape, otherwise a list of
        ``(1, z, y, x, channels)`` arrays will be returned.

    data_shape : List of tuples
        Shapes of all 3D images readed. Useful to reconstruct the original images together with ``crop_shape``.

    crop_shape : List of tuples
        Shape of the loaded 3D images after cropping. Useful to reconstruct the original images together with
        ``data_shape``.

    filenames : List of str, optional
        Loaded filenames.

    Examples
    --------
    ::

        # EXAMPLE 1
        # Case where we need to load 20 images of shape (1024, 1024, 91, 1)
        data_path = "data/train/x"

        data = load_data_from_dir(data_path)
        # The function will print list's first position array's shape. In this example:
        #     *** Loaded data[0] shape is (20, 91, 1024, 1024, 1)
        # Notice height, width and depth swap as skimage.io imread function
        # is used to load images

        # EXAMPLE 2
        # Same as example 1 but with unknown shape, cropping them into (256, 256, 40, 1) subvolumes with minimum
        # overlap and storing filenames.
        data_path = "data/train/x"

        X_test, orig_test_img_shapes, \
        crop_test_img_shapes, te_filenames = load_3d_images_from_dir(
            test_path, crop=True, crop_shape=(256, 256, 40, 1), overlap=(0,0,0), return_filenames=True)

        # The function will print the shape of the created array which its size is the concatenation in 0 axis of all
        # subvolumes created for each 3D image in the given path. For example:
        #     *** Loaded data shape is (350, 40, 256, 256, 1)
        # Notice height, width and depth swap as skimage.io imread function is used to load images.
    """
    if crop and crop_shape is None:
        raise ValueError("'crop_shape' must be provided when 'crop' is True")
    
    if preprocess_f != None and preprocess_cfg == None:
        raise ValueError("The preprocessing configuration ('preprocess_cfg') is missing.")

    print("Loading data from {}".format(data_dir))
    ids = sorted(next(os.walk(data_dir))[2])

    if len(ids) == 0:
        raise ValueError("No images found in dir {}".format(data_dir))

    if crop:
        from biapy.data.data_3D_manipulation import crop_3D_data_with_overlap

    data = []
    data_shape = []
    c_shape = []
    filenames = []
    ax = None

    # Read images
    for n, id_ in tqdm(enumerate(ids), total=len(ids)):
        if id_.endswith('.npy'):
            img = np.load(os.path.join(data_dir, id_))
        elif id_.endswith('.hdf5') or id_.endswith('.h5'):
            img = h5py.File(os.path.join(data_dir, id_),'r')
            img = np.array(img[list(img)[0]])
        else:
            img = imread(os.path.join(data_dir, id_))
        img = np.squeeze(img)

        if img.ndim < 3:
            raise ValueError("Read image seems to be 2D: {}. Path: {}".format(img.shape, os.path.join(data_dir, id_)))

        if img.ndim == 3: 
            img = np.expand_dims(img, -1)
        else:
            min_val = min(img.shape)
            channel_pos = img.shape.index(min_val)
            if channel_pos != 3 and img.shape[channel_pos] <= 4:
                new_pos = [x for x in range(4) if x != channel_pos]+[channel_pos,]
                img = img.transpose(new_pos)

        filenames.append(id_)
        if reflect_to_complete_shape: img = pad_and_reflect(img, crop_shape, verbose=verbose)
        
        if crop_shape is not None and check_channel:
            if crop_shape[-1] != img.shape[-1]:
                if crop_shape[-1] == 3 and convert_to_rgb:
                    img = np.repeat(img, 3, axis=-1)
                else:
                    raise ValueError("Channel of the patch size given {} does not correspond with the loaded image {}. "
                        "Please, check the channels of the images!".format(crop_shape[-1], img.shape[-1]))

        if preprocess_f == None:
            data_shape.append(img.shape)
            if crop and img.shape != crop_shape[:3]+(img.shape[-1],):
                img = crop_3D_data_with_overlap(img, crop_shape[:3]+(img.shape[-1],), overlap=overlap, padding=padding,
                                                median_padding=median_padding, verbose=verbose)
            else:
                img = np.expand_dims(img, axis=0)
            c_shape.append(img.shape)
        data.append(img)
    
    if preprocess_f != None:
        if is_mask:
            # data contains masks
            data = preprocess_f(preprocess_cfg, y_data = data, is_2d = False, is_y_mask = is_mask)
        else:
            data = preprocess_f(preprocess_cfg, x_data = data, is_2d = False)
    
        _data = []
        for img_num in range(len(data)):
            img = data[img_num]
            data_shape.append(img.shape)
            if crop and img.shape != crop_shape[:3]+(img.shape[-1],):
                img = crop_3D_data_with_overlap(img, crop_shape[:3]+(img.shape[-1],), overlap=overlap, padding=padding,
                                                median_padding=median_padding, verbose=verbose)
            else:
                img = np.expand_dims(img, axis=0)
            c_shape.append(img.shape)
            _data.append(img)
        del data
        data = _data

    same_shape = True
    s = data[0].shape
    dtype = data[0].dtype
    drange = data_range(data[0])
    for i in range(1,len(data)):
        if check_drange and drange != data_range(data[i]):
            raise ValueError("Input images ({} vs {}) seem to have different data ranges ({} and {} found) Please check it "
                "and ensure all images have same data type"
                .format(filenames[0], filenames[i], drange, data_range(data[i])))
        if s != data[i].shape:
            same_shape = False

    if crop or same_shape:
        data = np.concatenate(data)
        print("*** Loaded data shape is {}".format(data.shape))
    else:
        print("Not all samples seem to have the same shape. Number of samples: {}".format(len(data)))
        print("*** First sample shape is {}".format(data[0].shape))

    if return_filenames:
        return data, data_shape, c_shape, filenames
    else:
        return data, data_shape, c_shape


def check_downsample_division(X, d_levels):
    """Ensures ``X`` shape is divisible by ``2`` times ``d_levels`` adding padding if necessary.

       Parameters
       ----------
       X : 4D Numpy array
           Data to check if its shape.  E.g. ``(10, 1000, 1000, 1)``.

       d_levels : int
           Levels of downsampling by ``2``.

       Returns
       -------
       X : 4D Numpy array
           Data divisible by 2 ``d_levels`` times.

       o_shape : 4 int tuple
           Original shape of ``X``. E.g. ``(10, 1000, 1000, 1)``.
    """

    d_val = pow(2,d_levels)
    dy = math.ceil(X.shape[1]/d_val)
    dx = math.ceil(X.shape[2]/d_val)
    o_shape = X.shape
    if dy*d_val != X.shape[1] or dx*d_val != X.shape[2]:
        X = np.pad(X, ((0,0), (0,(dy*d_val)-X.shape[1]), (0,(dx*d_val)-X.shape[2]), (0,0)))
        print("Data has been padded to be downsampled {} times. Its shape now is: {}".format(d_levels, X.shape))
    return X, o_shape


def save_npy_files(X, data_dir=None, filenames=None, verbose=True):
    """Save images in the given directory.

       Parameters
       ----------
       X : 4D/5D numpy array
           Data to save as images. The first dimension must be the number of images. E.g.
           ``(num_of_images, y, x, channels)`` or ``(num_of_images, z, y, x, channels)``.

       data_dir : str, optional
           Path to store X images.

       filenames : List, optional
           Filenames that should be used when saving each image.

       verbose : bool, optional
            To print saving information.
    """

    if verbose:
        s = X.shape if not isinstance(X, list) else X[0].shape
        print("Saving {} data as .npy in folder: {}".format(s, data_dir))

    os.makedirs(data_dir, exist_ok=True)
    if filenames is not None:
        if len(filenames) != len(X):
            raise ValueError("Filenames array and length of X have different shapes: {} vs {}".format(len(filenames),len(X)))

    d = len(str(len(X)))
    for i in tqdm(range(len(X)), leave=False):
        if filenames is None:
            f = os.path.join(data_dir, str(i).zfill(d)+'.npy')
        else:
            f = os.path.join(data_dir, os.path.splitext(filenames[i])[0]+'.npy')
        if isinstance(X, list):
            np.save(f, X[i][0])
        else:
            np.save(f, X[i])

def pad_and_reflect(img, crop_shape, verbose=False):
    """Load data from a directory.

       Parameters
       ----------
       img : 3D/4D Numpy array
           Image to pad. E.g. ``(y, x, channels)`` or ``(z, y, x, c)``.

       crop_shape : Tuple of 3/4 ints, optional
           Shape of the subvolumes to create when cropping.  E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

       verbose : bool, optional
           Whether to output information.

       Returns
       -------
       img : 3D/4D Numpy array
           Image padded. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.
    """
    if img.ndim == 4 and len(crop_shape) != 4:
        raise ValueError("'crop_shape' needs to have 4 values as the input array has 4 dims")
    if img.ndim == 3 and len(crop_shape) != 3:
        raise ValueError("'crop_shape' needs to have 3 values as the input array has 3 dims")

    if img.ndim == 4:
        if img.shape[0] < crop_shape[0]:
            diff = crop_shape[0]-img.shape[0]
            o_shape = img.shape
            img = np.pad(img, ((diff,0),(0,0),(0,0),(0,0)), 'reflect')
            if verbose: print("Reflected from {} to {}".format(o_shape, img.shape))

        if img.shape[1] < crop_shape[1]:
            diff = crop_shape[1]-img.shape[1]
            o_shape = img.shape
            img = np.pad(img, ((0,0),(diff,0),(0,0),(0,0)), 'reflect')
            if verbose: print("Reflected from {} to {}".format(o_shape, img.shape))

        if img.shape[2] < crop_shape[2]:
            diff = crop_shape[2]-img.shape[2]
            o_shape = img.shape
            img = np.pad(img, ((0,0),(0,0),(diff,0),(0,0)), 'reflect')
            if verbose: print("Reflected from {} to {}".format(o_shape, img.shape))
    else:
        if img.shape[0] < crop_shape[0]:
            diff = crop_shape[0]-img.shape[0]
            o_shape = img.shape
            img = np.pad(img, ((diff,0),(0,0),(0,0)), 'reflect')
            if verbose: print("Reflected from {} to {}".format(o_shape, img.shape))

        if img.shape[1] < crop_shape[1]:
            diff = crop_shape[1]-img.shape[1]
            o_shape = img.shape
            img = np.pad(img, ((0,0),(diff,0),(0,0)), 'reflect')
            if verbose: print("Reflected from {} to {}".format(o_shape, img.shape))
    return img

    
def check_value(value, value_range=(0,1)):
    """Checks if a value is within a range """
    if isinstance(value, list) or isinstance(value, tuple):
        for i in range(len(value)):
            if isinstance(value[i], np.ndarray):
                if value_range[0] <= np.min(value[i]) or np.max(value[i]) <= value_range[1]:
                    return False
            else:
                if not (value_range[0] <= value[i] <= value_range[1]):    
                    return False
        return True 
    else:
        if isinstance(value, np.ndarray):
            if value_range[0] <= np.min(value) and np.max(value) <= value_range[1]:
                return True
        else:  
            if value_range[0] <= value <= value_range[1]:
                return True
        return False

def data_range(x):
    if not isinstance(x, np.ndarray):
        raise ValueError("Input array of type {} and not numpy array".format(type(x)))
    if check_value(x, (0,1)):
        return "01 range"
    elif check_value(x, (0,255)):
        return "uint8 range"
    elif check_value(x, (0,65535)):
        return "uint16 range"
    else:
        return "none_range"

def read_chunked_data(filename):
    if isinstance(filename, str):
        if filename.endswith('.hdf5') or filename.endswith('.h5'):
            fid = h5py.File(filename,'r')
            data = fid[list(fid)[0]]
        elif filename.endswith('.zarr'):
            fid = zarr.open(filename,'r')
            if len(list((fid.group_keys()))) != 0: # if the zarr has groups
                fid = fid[list(fid.group_keys())[0]]
            if len(list((fid.array_keys()))) != 0: # if the zarr has arrays
                data = fid[list(fid.array_keys())[0]]
            else: 
                data = fid
        else:
            raise ValueError(f"File extension {os.path.splitext(fid)[1]} not recognized")

        return fid, data

def write_chunked_data(data, data_dir, filename, dtype_str="float32", verbose=True):
    """
    Save images in the given directory.

    Parameters
    ----------
    X : 4D numpy array
        Data to save. E.g. ``(z, y, x, channels)``.

    data_dir : str
        Path to store X images.

    filename : str
        Filename of the data to use.

    dtype_str : str, optional
        Data type to use when saving. 

    verbose : bool, optional
        To print saving information.
    """
    if data.ndim != 4:
        raise ValueError(f"Expected data needs to have 4 dimensions. Given data shape: {data.shape}")

    ext = os.path.splitext(filename)[1]
    if verbose:
        print("Saving {} data as {} in folder: {}".format(data.shape, ext, data_dir))

    os.makedirs(data_dir, exist_ok=True)

    if ext in ['.hdf5', '.h5']:
        fid = h5py.File(os.path.join(data_dir, filename), "w") 
        data = fid.create_dataset("data", data.shape, dtype=dtype_str, compression="gzip")
    # Zarr
    else:
        fid = zarr.open_group(os.path.join(data_dir, filename), mode="w")
        data = fid.create_dataset("data", shape=data.shape, dtype=dtype_str)

def order_dimensions(data, input_order, output_order='TZCYX', default_value=1):
    """
    Reorder data from any input order to output order.

    Parameters
    ----------
    data : Numpy array like
        data to reorder. E.g. ``(z, y, x, channels)``.

    input_order : str
        Order of the input data. E.g. ``ZYXC``.

    output_order : str, optional
        Order of the output data. E.g. ``TZCYX``.

    default_value : Any, optional
        Default value to use when a dimension is not present in the input order.

    Returns
    -------
    shape : Tuple
        Reordered data. E.g. ``(t, z, channel, y, x)``.
    """

    if input_order == output_order:
        return data

    output_data = []

    for i in range(len(output_order)):
        if output_order[i] in input_order:
            output_data.append(data[input_order.index(output_order[i])])
        else:
            output_data.append(default_value)
    return tuple(output_data)


import os
import sys
import argparse
import datetime
import ntpath
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
from shutil import copyfile
import numpy as np
import importlib
import multiprocessing

from biapy.utils.misc import init_devices, is_dist_avail_and_initialized, set_seed
from biapy.config.config import Config
from biapy.engine.check_configuration import check_configuration

class BiaPy():
    def __init__(self, config, result_dir=os.getenv('HOME'), name="unknown_job", run_id=1, gpu=None, world_size=1, 
        local_rank=-1, dist_on_itp=False, dist_url='env://', dist_backend='nccl'):
        """
        Run the main functionality of the job.

        Parameters
        ----------
        config: str,  optional
            Path to the configuration file.

        result_dir: str, optional
            Path to where the resulting output of the job will be stored. Defaults to the home directory.
        
        name: str, optional
            Job name. Defaults to "unknown_job".

        run_id: int, optional
            Run number of the same job. Defaults to 1.

        gpu: str, optional
            GPU number according to 'nvidia-smi' command. Defaults to None.

        world_size: int, optional
            Number of distributed processes. Defaults to 1.

        local_rank: int, optional
            Node rank for distributed training. Necessary for using the torch.distributed.launch utility. Defaults to -1.

        dist_on_itp: bool, optional
            If True, distributed training is performed. Defaults to False.

        dist_url: str, optional
            URL used to set up distributed training. Defaults to 'env://'.

        dist_backend: str, optional
            Backend to use in distributed mode. Should be either 'nccl' or 'gloo'. Defaults to 'nccl'.
        """

        if dist_backend not in ['nccl', 'gloo']:
            raise ValueError("Invalid value for 'dist_backend'. Should be either 'nccl' or 'gloo'.")
    
        self.args = argparse.Namespace(
            config=config,
            result_dir=result_dir,
            name=name,
            run_id=run_id,
            gpu=gpu,
            world_size=world_size,
            local_rank=local_rank,
            dist_on_itp=dist_on_itp,
            dist_url=dist_url,
            dist_backend=dist_backend
        )
        
        ############
        #  CHECKS  #
        ############

        # Job complete name
        self.job_identifier = self.args.name + '_' + str(self.args.run_id)

        # Prepare working dir
        self.job_dir = os.path.join(self.args.result_dir, self.args.name)
        self.cfg_bck_dir = os.path.join(self.job_dir, 'config_files')
        os.makedirs(self.cfg_bck_dir, exist_ok=True)
        head, tail = ntpath.split(self.args.config)
        self.cfg_filename = tail if tail else ntpath.basename(head)
        self.cfg_file = os.path.join(self.cfg_bck_dir,self.cfg_filename)

        if not os.path.exists(self.args.config):
            raise FileNotFoundError("Provided {} config file does not exist".format(self.args.config))
        copyfile(self.args.config, self.cfg_file)

        # Merge conf file with the default settings
        self.cfg = Config(self.job_dir, self.job_identifier)
        self.cfg._C.merge_from_file(self.cfg_file)
        self.cfg.update_dependencies()
        #self.cfg.freeze()

        now = datetime.datetime.now()
        print("Date: {}".format(now.strftime("%Y-%m-%d %H:%M:%S")))
        print("Arguments: {}".format(self.args))
        print("Job: {}".format(self.job_identifier))
        print("Python       : {}".format(sys.version.split('\n')[0]))
        print("PyTorch: ", torch.__version__)

        # GPU selection
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        opts = []
        if self.args.gpu:
            os.environ["CUDA_VISIBLE_DEVICES"] = self.args.gpu
            self.num_gpus = len(np.unique(np.array(self.args.gpu.strip().split(','))))
            opts.extend(["SYSTEM.NUM_GPUS", self.num_gpus])

        # GPU management
        self.device = init_devices(self.args, self.cfg.get_cfg_defaults())
        self.cfg._C.merge_from_list(opts)
        self.cfg = self.cfg.get_cfg_defaults()

        # Reproducibility
        set_seed(self.cfg.SYSTEM.SEED)
        
        # Number of CPU calculation
        if self.cfg.SYSTEM.NUM_CPUS == -1:
            self.cpu_count = multiprocessing.cpu_count()
        else:
            self.cpu_count = self.cfg.SYSTEM.NUM_CPUS
        if self.cpu_count < 1: self.cpu_count = 1 # At least 1 CPU
        torch.set_num_threads(self.cpu_count)
        self.cfg.merge_from_list(['SYSTEM.NUM_CPUS', self.cpu_count])

        check_configuration(self.cfg, self.job_identifier)
        print("Configuration details:")
        print(self.cfg)

        ##########################
        #       TRAIN/TEST       #
        ##########################
        workflowname = str(self.cfg.PROBLEM.TYPE).lower()
        mdl = importlib.import_module('biapy.engine.'+workflowname)
        names = [x for x in mdl.__dict__ if not x.startswith("_")]
        globals().update({k: getattr(mdl, k) for k in names})
        name = [x for x in names if "Base" not in x and "Workflow" in x][0]

        # Initialize workflow
        print("*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*")
        print(f"Initializing {name}")
        print("*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*\n")
        self.workflow = getattr(mdl, name)(self.cfg, self.job_identifier, self.device, self.args)

    def train(self):
        """Call training phase."""
        if self.cfg.TRAIN.ENABLE:
            self.workflow.train()
        else:
            raise ValueError("Train was not enabled ('TRAIN.ENABLE')")

    def test(self):
        """Call test phase."""
        if self.cfg.TEST.ENABLE:
            self.workflow.test()
        else:
            raise ValueError("Test was not enabled ('TEST.ENABLE')")

    def prepare_model(self):
        """Build up the model based on the selected configuration."""
        self.workflow.prepare_model()

    def export_model_to_bmz(self, bmz_cfg):
        """
        Export a model into Bioimage Model Zoo format. 

        Parameters
        ----------
        bmz_cfg : BMZ configuration
            BMZ configuration to export the model. Here multiple keys need to be declared:

            description : str
                Description of the model.

            authors : list of dicts
                Authors of the model. Need to be a list of dicts. E.g. ``[{"name": "Gizmo"}]``. 

            license : str
                License of the model. E.g. "CC-BY-4.0"

            tags : List of str
                Tags to make models more findable on the website. E.g. ``["nucleus-segmentation"]``.

            cite : List of dicts 
                List of dictionaries of citations associated. E.g. 
                ``[{"text": "Gizmo et al.", "doi": "doi:10.1002/xyzacab123"}]``

            doc : path
                Path to a file with a documentation of the model in markdown. E.g. "my-model/doc.md"
                
            build_dir : path
                Path to store files and the build of BMZ package. 

            model_name : str, optional
                Name of the model. If not set a name based on the selected configuration
                will be created. 

            input_axes : List of str, optional
                Axis order of the input file. E.g. ["bcyx"].
            
            output_axes : List of str, optional
                Axis order of the output file. E.g. ["bcyx"].

            test_input : 3D/4D Torch tensor, optional
                Test input image sample. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.

            test_output : 3D/4D Torch tensor, optional
                Test output image sample. E.g. ``(y, x, channels)`` or ``(z, y, x, channels)``.
        """
        # Check keys
        need_info = ['description', 'authors', 'license', 'tags', 'cite', 'doc', 'build_dir']
        for x in need_info:
            if x not in bmz_cfg:
                raise ValueError(f"'{x}' property must be declared in 'bmz_cfg'")

        # Check if BiaPy has been run so some of the variables have been created
        if not self.workflow.model_prepared:
            raise ValueError("You need first to call prepare_model(), train(), test() or run_job() functions so the model can be built")
        error = False
        if self.workflow.bmz_test_input is None:
            if 'test_input' not in bmz_cfg:
                error = True
            elif bmz_cfg['test_input'] is None:
                error = True
            if error:
                raise ValueError("No bmz_cfg['test_input'] available. You can: 1) provide it using bmz_config['test_input'] "
                    "or run the training phase, by calling train() or run_job() functions.")
        if self.workflow.bmz_test_output is None:
            if 'test_output' not in bmz_cfg:
                error = True
            elif bmz_cfg['test_output'] is None:
                error = True
            if error:
                raise ValueError("No bmz_cfg['test_output'] available. You can: 1) provide it using bmz_config['test_output'] "
                    "or run the training phase, by calling train() or run_job() functions.")

        # Check BMZ dictionary keys and values 
        if bmz_cfg['description'] == "":
            raise ValueError("'bmz_cfg['description']' can not be empty.")
        if not isinstance(bmz_cfg['authors'], list):
            raise ValueError("'bmz_cfg['authors']' needs to be a list of dicts. E.g. [{'name': 'Daniel'}]")
        else:
            if len(bmz_cfg['authors']) == 0:
                raise ValueError("'bmz_cfg['authors']' can not be empty.")
            for d in bmz_cfg['authors']:
                if not isinstance(d, dict):
                    raise ValueError("'bmz_cfg['authors']' must be a list of dicts. E.g. [{'name': 'Daniel'}]")
        if bmz_cfg['license'] == "":
                raise ValueError("'bmz_cfg['license']' can not be empty. E.g. 'CC-BY-4.0'")
        if not isinstance(bmz_cfg['tags'], list):
            raise ValueError("'bmz_cfg['tags']' needs to be a list of dicts. E.g. [{'modality': 'electron-microscopy', 'content': 'mitochondria'}]")
        else:
            if len(bmz_cfg['tags']) == 0:
                raise ValueError("'bmz_cfg['tags']' can not be empty")
            for d in bmz_cfg['tags']:
                if not isinstance(d, dict):
                    raise ValueError("'bmz_cfg['tags']' must be a list of dicts. E.g. [{'modality': 'electron-microscopy', 'content': 'mitochondria'}]")                    
        if not isinstance(bmz_cfg['cite'], list):
            raise ValueError("'bmz_cfg['cite']' needs to be a list of dicts. E.g. [{'text': 'Gizmo et al.', 'doi': 'doi:10.1002/xyzacab123'}]")
        else:
            for d in bmz_cfg['cite']:
                if not isinstance(d, dict):
                    raise ValueError("'bmz_cfg['cite']' needs to be a list of dicts. E.g. [{'text': 'Gizmo et al.', 'doi': 'doi:10.1002/xyzacab123'}]")
        if bmz_cfg['doc'] == "":
            raise ValueError("'bmz_cfg['doc']' can not be empty. E.g. '/home/user/my-model/doc.md'")
        else:
            if not os.path.exists(bmz_cfg['doc']):
                raise ValueError("'bmz_cfg['doc']' file does not exist!")
        if bmz_cfg['build_dir'] == "":
            raise ValueError("'bmz_cfg['build_dir']' can not be empty.")

        if 'input_axes' in bmz_cfg:
            if not isinstance(bmz_cfg['input_axes'], list):
                raise ValueError("'bmz_cfg['input_axes']' needs to be a list containing just one str. E.g. ['bcyx'].")
            if len(bmz_cfg['input_axes']) != 1: 
                raise ValueError("'bmz_cfg['input_axes']' needs to be a list containing just one str. E.g. ['bcyx'].")  
            if not isinstance(bmz_cfg['input_axes'][0], str): 
                raise ValueError("'bmz_cfg['input_axes']' needs to be a list containing just one str. E.g. ['bcyx'].") 
        if 'output_axes' in bmz_cfg:
            if not isinstance(bmz_cfg['output_axes'], list):
                raise ValueError("'bmz_cfg['output_axes']' needs to be a list containing just one str. E.g. ['bcyx'].")
            if len(bmz_cfg['output_axes']) != 1: 
                raise ValueError("'bmz_cfg['output_axes']' needs to be a list containing just one str. E.g. ['bcyx'].")  
            if not isinstance(bmz_cfg['output_axes'][0], str): 
                raise ValueError("'bmz_cfg['output_axes']' needs to be a list containing just one str. E.g. ['bcyx'].") 
        if 'test_input' in bmz_cfg and not torch.is_tensor(bmz_cfg['test_input']):
            raise ValueError("'bmz_cfg['test_input']' needs to be a Tensor") 
        if 'test_output' in bmz_cfg and not torch.is_tensor(bmz_cfg['test_output']):
            raise ValueError("'bmz_cfg['test_output']' needs to be a Tensor") 

        from bioimageio.core.build_spec import build_model
        from bioimageio.core.resource_tests import test_model
        from bioimageio.core import load_resource_description

        # Save input/output samples
        os.makedirs(bmz_cfg['build_dir'], exist_ok=True)
        input_sample_path = os.path.join(bmz_cfg['build_dir'], "test-input.npy")
        output_sample_path = os.path.join(bmz_cfg['build_dir'], "test-output.npy")
        test_input = self.workflow.bmz_test_input if 'test_input' not in bmz_cfg else bmz_cfg['test_input']
        test_output = self.workflow.bmz_test_output if 'test_output' not in bmz_cfg else bmz_cfg['test_output']
        
        if test_input.ndim == 3:
            np.save(input_sample_path, test_input.permute((2, 0 ,1)).unsqueeze(0))
            in_axis = ["bcyx"]
        else:
            np.save(input_sample_path, test_input.permute((3, 0 ,1, 2)).unsqueeze(0))
            in_axis = ["bczyx"]

        if test_output.ndim == 3:
            np.save(output_sample_path, test_output.permute((2, 0 ,1)).unsqueeze(0))
            out_axis = ["bcyx"]
        elif test_output.ndim == 4:
            np.save(output_sample_path, test_output.permute((3, 0 ,1, 2)).unsqueeze(0))
            out_axis = ["bczyx"]
        else:
            np.save(output_sample_path, test_output.unsqueeze(0))
            out_axis = ["bc"]

        # Name of the model
        if 'model_name' in bmz_cfg:
            model_name = bmz_cfg['model_name']
        else:
            if self.cfg.MODEL.SOURCE == "biapy":
                model_name = "my_"+self.cfg.MODEL.ARCHITECTURE+"_"+self.cfg.PROBLEM.NDIM
            elif self.cfg.MODEL.SOURCE == "torchvision":
                model_name = "my_"+self.cfg.MODEL.TORCHVISION_MODEL_NAME+"_"+self.cfg.PROBLEM.NDIM
            else:
                model_name = "my_BMZ_"+self.cfg.MODEL.BMZ.SOURCE_MODEL_DOI

        # Preprocessing
        # Actually Torchvision has its own preprocessing but it can not be adapted to BMZ easily, so for now
        # we set it like we were using BiaPy backend
        if self.cfg.MODEL.SOURCE in ["biapy", "torchvision"]:
            if self.cfg.DATA.NORMALIZATION.TYPE == 'div':
                preprocessing = [[{"name": "scale_linear", "kwargs": {"gain": 1/255, "offset": 0}}]]
            else:
                if not os.path.exists(self.cfg.PATHS.MEAN_INFO_FILE) or not os.path.exists(self.cfg.PATHS.STD_INFO_FILE):
                    raise FileNotFoundError("Not mean/std files found in {} and {}"
                        .format(self.cfg.PATHS.MEAN_INFO_FILE, self.cfg.PATHS.STD_INFO_FILE))
                custom_mean = float(np.load(self.cfg.PATHS.MEAN_INFO_FILE))
                custom_std = float(np.load(self.cfg.PATHS.STD_INFO_FILE))
                preprocessing = [[{"name": "zero_mean_unit_variance", "kwargs": {"mean": custom_mean, "std": custom_std, 'mode': 'per_sample'}}]]
        else: # BMZ
            # Done as in https://github.com/bioimage-io/core-bioimage-io-python/blob/main/example/model_creation.ipynb
            preprocessing = [[{"name": prep.name, "kwargs": prep.kwargs} for prep in inp.preprocessing] for inp in self.workflow.bmz_model_resource.inputs]

        # Post-processing (not clear for now so just output the raw output of the model)
        postprocessing = None
        # if self.cfg.PROBLEM.TYPE in ['SEMANTIC_SEG', 'DETECTION', "SUPER_RESOLUTION", "SELF_SUPERVISED"]:
        #     postprocessing = [{"name": "binarize", "kwargs": {"threshold": 0.5}}]

        # Configure tags. For now there is no "biapy" tag so we are filling as much as we can
        tags = bmz_cfg['tags'][0]
        if 'dims' not in tags:
            tags['dims'] = self.cfg.PROBLEM.NDIM.lower()
        tags['framework'] = "pytorch"
        # tags['software'] = "biapy" # {'software': ['ilastik', 'imagej', 'fiji', 'imjoy', 'deepimagej', 'napari']}
        # tags['method'] = # {'method': ['stardist', 'cellpose', 'yolo', 'care', 'n2v', 'denoiseg']},
        # tags['task'] = {'task': ['semantic-segmentation', 'instance-segmentation', 'object-detection', 'image-classification', \
        # 'denoising', 'image-restoration', 'image-reconstruction', 'in-silico-labeling']}

        # Change dir as the building process copies to the current directory the files used to create the BMZ model
        cwd = os.getcwd()
        os.chdir(bmz_cfg['build_dir'])

        # Save model's weights 
        torchscript_model = torch.jit.script(self.workflow.model)
        weight_file = os.path.join(bmz_cfg['build_dir'], "weights.pt")
        torchscript_model.save(weight_file)

        # Export model to BMZ format 
        build_model(
            # the weight file and the type of the weights
            weight_uri=weight_file,
            weight_type="torchscript",
            pytorch_version=str(torch.__version__),
            # the test input and output data as well as the description of the tensors
            # these are passed as list because we support multiple inputs / outputs per model
            test_inputs=[input_sample_path],
            test_outputs=[output_sample_path],
            input_axes=in_axis if 'input_axes' not in bmz_cfg else bmz_cfg['input_axes'],
            output_axes=out_axis if 'output_axes' not in bmz_cfg else bmz_cfg['output_axes'],
            # where to save the model zip, how to call the model and a short description of it
            output_path=os.path.join(bmz_cfg['build_dir'], "model.zip"),
            name=model_name,
            description=bmz_cfg['description'],
            # additional metadata about authors, licenses, citation etc.
            authors=bmz_cfg['authors'],
            license=bmz_cfg['license'],
            documentation=bmz_cfg['doc'],
            tags=[tags],
            cite=bmz_cfg['cite'],
            preprocessing=preprocessing,
            postprocessing=postprocessing,
        )

        # Checking model consistency
        my_model = load_resource_description(os.path.join(bmz_cfg['build_dir'], "model.zip") )
        test_model(my_model)

        # Recover the original working path
        os.chdir(cwd)


    def run_job(self):
        """Run a complete BiaPy workflow."""
        if self.cfg.TRAIN.ENABLE:
            self.train()

        if self.cfg.TEST.ENABLE:
            self.test()

        if is_dist_avail_and_initialized():
            dist.barrier()
            dist.destroy_process_group()

        print("FINISHED JOB {} !!".format(self.job_identifier))